"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""

from PKDevTools.classes.ColorText import colorText

from pkscreener.classes.Pktalib import pktalib
# from PKDevTools.classes.log import measure_time

class CandlePatterns:
    reversalPatternsBullish = [
        "Morning Star",
        "Morning Doji Star",
        "3 Inside Up",
        "Hammer",
        "3 White Soldiers",
        "Bullish Engulfing",
        "Dragonfly Doji",
        "Supply Drought",
        "Demand Rise",
        "Cup and Handle",
    ]
    reversalPatternsBearish = [
        "Evening Star",
        "Evening Doji Star",
        "3 Inside Down",
        "Inverted Hammer",
        "Hanging Man",
        "3 Black Crows",
        "Bearish Engulfing",
        "Shooting Star",
        "Gravestone Doji",
    ]

    def __init__(self):
        pass

    def findCurrentSavedValue(self, screenDict, saveDict, key):
        existingScreen = screenDict.get(key)
        existingSave = saveDict.get(key)
        existingScreen = f"{existingScreen}, " if (existingScreen is not None and len(existingScreen) > 0) else ""
        existingSave = f"{existingSave}, " if (existingSave is not None and len(existingSave) > 0) else ""
        return existingScreen, existingSave

    #@measure_time
    # Find candle-stick patterns
    # Arrange if statements with max priority from top to bottom
    def findPattern(self, processedData, dict, saveDict,filterPattern=None):
        data = processedData.head(4)
        data = data[::-1]
        hasCandleStickPattern = False
        if "Pattern" not in saveDict.keys():
            saveDict["Pattern"] = ""
            dict["Pattern"] = ""
        # Only 'doji' and 'inside' is internally implemented by pandas_ta_classic.
        # Otherwise, for the rest of the candle patterns, they also need
        # TA-Lib.
        check = pktalib.CDLDOJI(data["open"], data["high"], data["low"], data["close"])
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + colorText.GREEN + f"Doji" + colorText.END 
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Doji"
            hasCandleStickPattern = True

        check = pktalib.CDLMORNINGSTAR(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"Morning Star" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Morning Star"
            hasCandleStickPattern = True
        
        check = pktalib.CDLCUPANDHANDLE(
            processedData["open"], processedData["high"], processedData["low"], processedData["close"]
        )
        if check:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"Cup and Handle" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Cup and Handle"
            hasCandleStickPattern = True

        check = pktalib.CDLMORNINGDOJISTAR(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"Morning Doji Star" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Morning Doji Star"
            hasCandleStickPattern = True

        check = pktalib.CDLEVENINGSTAR(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.FAIL + f"Evening Star" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Evening Star"
            hasCandleStickPattern = True

        check = pktalib.CDLEVENINGDOJISTAR(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.FAIL + f"Evening Doji Star" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Evening Doji Star"
            hasCandleStickPattern = True

        check = pktalib.CDLLADDERBOTTOM(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN + f"Bullish Ladder Bottom" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bullish Ladder Bottom"
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL + f"Bearish Ladder Bottom" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bearish Ladder Bottom"
            hasCandleStickPattern = True

        check = pktalib.CDL3LINESTRIKE(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN + f"3 Line Strike" + colorText.END 
                )
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL + f"3 Line Strike" + colorText.END 
                )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 Line Strike"
            hasCandleStickPattern = True

        check = pktalib.CDL3BLACKCROWS(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.FAIL + f"3 Black Crows" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 Black Crows"
            hasCandleStickPattern = True

        check = pktalib.CDL3INSIDE(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN + f"3 Inside Up" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 Outside Up"
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL + f"3 Inside Down" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 Inside Down"
            hasCandleStickPattern = True

        check = pktalib.CDL3OUTSIDE(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN + f"3 Outside Up" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 Outside Up"
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL + f"3 Outside Down" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 Outside Down"
            hasCandleStickPattern = True

        check = pktalib.CDL3WHITESOLDIERS(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"3 White Soldiers" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"3 White Soldiers"
            hasCandleStickPattern = True

        check = pktalib.CDLHARAMI(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN + f"Bullish Harami" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bullish Harami"
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL + f"Bearish Harami" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bearish Harami"
            hasCandleStickPattern = True

        check = pktalib.CDLHARAMICROSS(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN
                    + f"Bullish Harami Cross" 
                    + colorText.END
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bullish Harami Cross"
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL
                    + f"Bearish Harami Cross" 
                    + colorText.END
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bearish Harami Cross"
            hasCandleStickPattern = True

        check = pktalib.CDLMARUBOZU(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN
                    + f"Bullish Marubozu" 
                    + colorText.END
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bullish Marubozu"
            else:
                dict["Pattern"] = (
                    colorText.FAIL + f"Bearish Marubozu" + colorText.END 
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bearish Marubozu"
            hasCandleStickPattern = True

        check = pktalib.CDLHANGINGMAN(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.FAIL + f"Hanging Man" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Hanging Man"
            hasCandleStickPattern = True

        check = pktalib.CDLHAMMER(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"Hammer" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Hammer"
            hasCandleStickPattern = True

        check = pktalib.CDLINVERTEDHAMMER(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"Inverted Hammer" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Inverted Hammer"
            hasCandleStickPattern = True

        check = pktalib.CDLSHOOTINGSTAR(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.FAIL + f"Shooting Star" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Shooting Star"
            hasCandleStickPattern = True

        check = pktalib.CDLDRAGONFLYDOJI(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.GREEN + f"Dragonfly Doji" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Dragonfly Doji"
            hasCandleStickPattern = True

        check = pktalib.CDLGRAVESTONEDOJI(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                colorText.FAIL + f"Gravestone Doji" + colorText.END 
            )
            saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Gravestone Doji"
            hasCandleStickPattern = True

        check = pktalib.CDLENGULFING(
            data["open"], data["high"], data["low"], data["close"]
        )
        if check is not None and check.tail(1).item() != 0:
            if check.tail(1).item() > 0:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.GREEN
                    + f"Bullish Engulfing" 
                    + colorText.END
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bullish Engulfing"
            else:
                dict["Pattern"] = (self.findCurrentSavedValue(dict,saveDict,"Pattern")[0] + 
                    colorText.FAIL
                    + f"Bearish Engulfing" 
                    + colorText.END
                )
                saveDict["Pattern"] = self.findCurrentSavedValue(dict,saveDict,"Pattern")[1] +  f"Bearish Engulfing"
            hasCandleStickPattern = True
        if hasCandleStickPattern:
            return filterPattern in saveDict["Pattern"] if filterPattern is not None else hasCandleStickPattern
        return False



#!/usr/bin/python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""

import os
import random
import warnings
warnings.simplefilter("ignore", UserWarning, append=True)
os.environ["PYTHONWARNINGS"] = "ignore::UserWarning"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
import logging
import multiprocessing
import sys
import time
import urllib
import warnings
from datetime import datetime, UTC, timedelta
from time import sleep

import numpy as np

warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)
import pandas as pd
from alive_progress import alive_bar
from PKDevTools.classes.Committer import Committer
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKDevTools.classes import Archiver
from PKDevTools.classes.Telegram import (
    is_token_telegram_configured,
    send_document,
    send_message,
    send_photo,
    send_media_group
)
from PKNSETools.morningstartools.PKMorningstarDataFetcher import morningstarDataFetcher
from PKNSETools.Nasdaq.PKNasdaqIndex import PKNasdaqIndexFetcher
from tabulate import tabulate
from halo import Halo

import pkscreener.classes.ConfigManager as ConfigManager
import pkscreener.classes.Fetcher as Fetcher
import pkscreener.classes.ScreeningStatistics as ScreeningStatistics
from pkscreener.classes import Utility, ConsoleUtility, ConsoleMenuUtility, ImageUtility
from pkscreener.classes.Utility import STD_ENCODING
from pkscreener.classes import VERSION, PortfolioXRay
from pkscreener.classes.Backtest import backtest, backtestSummary
from pkscreener.classes.PKSpreadsheets import PKSpreadsheets
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.Environment import PKEnvironment
from pkscreener.classes.CandlePatterns import CandlePatterns
from pkscreener.classes import AssetsManager
from PKDevTools.classes.FunctionTimeouts import exit_after
from pkscreener.classes.MenuOptions import (
    level0MenuDict,
    level1_X_MenuDict,
    level1_P_MenuDict,
    level2_X_MenuDict,
    level2_P_MenuDict,
    level3_X_ChartPattern_MenuDict,
    level3_X_PopularStocks_MenuDict,
    level3_X_PotentialProfitable_MenuDict,
    PRICE_CROSS_SMA_EMA_DIRECTION_MENUDICT,
    PRICE_CROSS_SMA_EMA_TYPE_MENUDICT,
    PRICE_CROSS_PIVOT_POINT_TYPE_MENUDICT,
    level3_X_Reversal_MenuDict,
    level4_X_Lorenzian_MenuDict,
    level4_X_ChartPattern_Confluence_MenuDict,
    level4_X_ChartPattern_BBands_SQZ_MenuDict,
    level4_X_ChartPattern_MASignalMenuDict,
    level1_index_options_sectoral,
    menus,
    MAX_SUPPORTED_MENU_OPTION,
    MAX_MENU_OPTION,
    PIPED_SCANNERS,
    PREDEFINED_SCAN_MENU_KEYS,
    PREDEFINED_SCAN_MENU_TEXTS,
    INDICES_MAP,
    CANDLESTICK_DICT
)
from pkscreener.classes.OtaUpdater import OTAUpdater
from pkscreener.classes.Portfolio import PortfolioCollection
from pkscreener.classes.PKTask import PKTask
from pkscreener.classes.PKScheduler import PKScheduler
from pkscreener.classes.PKScanRunner import PKScanRunner
from pkscreener.classes.PKMarketOpenCloseAnalyser import PKMarketOpenCloseAnalyser
from pkscreener.classes.PKPremiumHandler import PKPremiumHandler
from pkscreener.classes.AssetsManager import PKAssetsManager
from pkscreener.classes.PKAnalytics import PKAnalyticsService
from pkscreener.classes.MenuManager import MenuManager, ScanExecutor, ResultProcessor, TelegramNotifier, DataManager, BacktestManager
if __name__ == '__main__':
    multiprocessing.freeze_support()

# Constants
np.seterr(divide="ignore", invalid="ignore")
TEST_STKCODE = "SBIN"


class PKScreenerMain:
    """
    Main application class for PKScreener that orchestrates the entire screening process.
    Coordinates between all manager classes and handles the main execution flow.
    """

    def __init__(self):
        """Initialize the PKScreener application with all manager classes."""
        self.config_manager = ConfigManager.tools()
        self.config_manager.getConfig(ConfigManager.parser)
        self.user_passed_args = None
        self.default_answer = None
        
        # Initialize manager classes
        self.menu_manager = MenuManager(self.config_manager, self.user_passed_args)
        self.scan_executor = ScanExecutor(self.config_manager, self.user_passed_args)
        self.result_processor = ResultProcessor(self.config_manager, self.user_passed_args)
        self.telegram_notifier = TelegramNotifier()
        self.data_manager = DataManager(self.config_manager, self.user_passed_args)
        self.backtest_manager = BacktestManager(self.config_manager, self.user_passed_args)
        
        # Share state between managers
        self.menu_manager.list_stock_codes = self.data_manager.list_stock_codes
        self.scan_executor.selected_choice = self.menu_manager.selected_choice
        self.scan_executor.criteria_date_time = self.result_processor.criteria_date_time
        self.result_processor.selected_choice = self.menu_manager.selected_choice
        self.result_processor.menu_choice_hierarchy = self.menu_manager.menu_choice_hierarchy
        self.data_manager.selected_choice = self.menu_manager.selected_choice
        self.data_manager.default_answer = self.default_answer
        self.backtest_manager.selected_choice = self.menu_manager.selected_choice
        self.backtest_manager.menu_choice_hierarchy = self.menu_manager.menu_choice_hierarchy
        self.backtest_manager.elapsed_time = self.scan_executor.elapsed_time
        self.backtest_manager.default_answer = self.default_answer

    def main(self, userArgs=None, optionalFinalOutcome_df=None):
        """
        Main entry point for the PKScreener application.
        
        This method orchestrates the entire screening process by coordinating between
        all manager classes. It handles menu navigation, scanning execution, result
        processing, and notifications while maintaining all existing functionality.
        
        Args:
            userArgs: User arguments passed to the application
            optionalFinalOutcome_df: Optional final outcome dataframe for intraday analysis
            
        Returns:
            tuple: Screen results and save results dataframes
        """
        # Initialize state variables
        self.user_passed_args = userArgs
        self.default_answer = None if userArgs is None else userArgs.answerdefault
        
        # Update references with actual user args
        self.menu_manager.user_passed_args = self.user_passed_args
        self.scan_executor.user_passed_args = self.user_passed_args
        self.result_processor.user_passed_args = self.user_passed_args
        self.telegram_notifier.user_passed_args = self.user_passed_args
        self.data_manager.user_passed_args = self.user_passed_args
        self.backtest_manager.user_passed_args = self.user_passed_args
        
        # Set initial state
        testing = False if userArgs is None else (userArgs.testbuild and userArgs.prodbuild)
        testBuild = False if userArgs is None else (userArgs.testbuild and not testing)
        downloadOnly = False if userArgs is None else userArgs.download
        startupoptions = None if userArgs is None else userArgs.options
        user = None if userArgs is None else userArgs.user
        self.default_answer = None if userArgs is None else userArgs.answerdefault
        
        # Initialize screening counters
        self.scan_executor.screen_counter = multiprocessing.Value("i", 1)
        self.scan_executor.screen_results_counter = multiprocessing.Value("i", 0)
        
        # Initialize multiprocessing manager
        if self.scan_executor.mp_manager is None:
            self.scan_executor.mp_manager = multiprocessing.Manager()
            
        # Setup keyboard interrupt handling
        if self.scan_executor.keyboard_interrupt_event is None and not self.scan_executor.keyboard_interrupt_event_fired:
            self.scan_executor.keyboard_interrupt_event = self.scan_executor.mp_manager.Event()
            mkt_monitor_dict = self.scan_executor.mp_manager.dict()
            self.startMarketMonitor(mkt_monitor_dict, self.scan_executor.keyboard_interrupt_event)
            
        self.scan_executor.keyboard_interrupt_event_fired = False
        
        # Initialize stock data dictionaries
        if self.data_manager.stock_dict_primary is None or isinstance(self.data_manager.stock_dict_primary, dict):
            self.data_manager.stock_dict_primary = self.scan_executor.mp_manager.dict()
            self.data_manager.stock_dict_secondary = self.scan_executor.mp_manager.dict()
            self.data_manager.load_count = 0
            
        # Handle cleanup if needed
        if not self.data_manager.run_clean_up and self.user_passed_args is not None and not self.user_passed_args.systemlaunched:
            self.data_manager.cleanup_local_results()
            
        # Log user arguments if enabled
        if self.user_passed_args.log:
            default_logger().debug(f"User Passed args: {self.user_passed_args}")
            
        # Initialize results dataframes
        self.scan_executor.screen_results, self.scan_executor.save_results = PKScanRunner.initDataframes()
        
        # Get top level menu choices
        options, menuOption, indexOption, executeOption = self.menu_manager.getTopLevelMenuChoices(
            startupoptions, testBuild, downloadOnly, defaultAnswer=self.default_answer
        )
        
        # Execute main menu navigation and processing
        selectedMenu = self.menu_manager.initExecution(menuOption=menuOption)
        menuOption = selectedMenu.menuKey
        
        # Handle premium feature checks
        if menuOption in ["F", "M", "S", "B", "G", "C", "P", "D"] or selectedMenu.isPremium:
            self.menu_manager.ensureMenusLoaded(menuOption, indexOption, executeOption)
            if not PKPremiumHandler.hasPremium(selectedMenu):
                PKAnalyticsService().send_event(f"non_premium_user_{menuOption}_{indexOption}_{executeOption}")
                PKAnalyticsService().send_event("app_exit")
                sys.exit(0)
                
        # Handle special menu options
        if menuOption in ["M", "D", "I", "L", "F"]:
            self.handle_special_menu_options(menuOption)
            return None, None
            
        # Process scanner menu choices
        if menuOption in ["X", "T", "E", "Y", "U", "H", "C"]:
            menuOption, indexOption, executeOption, self.menu_manager.selected_choice = self.menu_manager.getScannerMenuChoices(
                testBuild or testing, downloadOnly, startupoptions, menuOption=menuOption,
                indexOption=indexOption, executeOption=executeOption, defaultAnswer=self.default_answer,
                user=user
            )
            if indexOption is None:
                return None, None
                
        # Handle backtest options
        elif menuOption in ["B", "G"]:
            indexOption, executeOption, backtestPeriod = self.backtest_manager.takeBacktestInputs(
                str(menuOption).upper(), indexOption, executeOption, 0
            )
            backtestPeriod = backtestPeriod * self.config_manager.backtestPeriodFactor
            
        # Handle strategy screening
        elif menuOption in ["S"]:
            strategyFilter = self.handle_strategy_screening(options)
            if strategyFilter:
                menuOption, indexOption, executeOption, self.menu_manager.selected_choice = self.menu_manager.getScannerMenuChoices(
                    testBuild or testing, downloadOnly, startupoptions, menuOption="X",
                    indexOption=indexOption, executeOption=executeOption, defaultAnswer=self.default_answer,
                    user=user
                )
                
        # Update menu choice hierarchy
        self.menu_manager.update_menu_choice_hierarchy()
        
        # Prepare stocks for screening
        self.data_manager.list_stock_codes = self.menu_manager.handleRequestForSpecificStocks(options, indexOption)
        self.data_manager.list_stock_codes = self.data_manager.prepareStocksForScreening(
            testing, downloadOnly, self.data_manager.list_stock_codes, indexOption
        )
        
        # Handle exit requests
        self.menu_manager.handleExitRequest(executeOption)
        
        # Process execute options
        volumeRatio = self.config_manager.volumeRatio
        reversalOption = None
        respChartPattern = None
        daysForLowestVolume = 30
        maLength = None
        
        if executeOption == 3:
            self.user_passed_args.maxdisplayresults = max(self.config_manager.maxdisplayresults, 2000)
        elif executeOption == 4:
            daysForLowestVolume = self.menu_manager.handleScannerExecuteOption4(executeOption, options)
        elif executeOption == 5:
            minRSI, maxRSI = ConsoleMenuUtility.PKConsoleMenuTools.promptRSIValues()
        elif executeOption == 6:
            reversalOption, maLength = ConsoleMenuUtility.PKConsoleMenuTools.promptReversalScreening(
                self.menu_manager.m2.find(str(executeOption))
            )
        elif executeOption == 7:
            respChartPattern, insideBarToLookback = ConsoleMenuUtility.PKConsoleMenuTools.promptChartPatterns(
                self.menu_manager.m2.find(str(executeOption))
            )
            if respChartPattern in [3, 6, 9]:
                maLength = ConsoleMenuUtility.PKConsoleMenuTools.promptChartPatternSubMenu(
                    self.menu_manager.m2.find(str(executeOption)), respChartPattern
                )
        # ... handle other execute options
        
        # Load or fetch stock data
        if not self.data_manager.loaded_stock_data:
            try:
                import tensorflow as tf
                with tf.device("/device:GPU:0"):
                    self.data_manager.stock_dict_primary, self.data_manager.stock_dict_secondary = self.data_manager.loadDatabaseOrFetch(
                        downloadOnly, self.data_manager.list_stock_codes, menuOption, indexOption
                    )
            except:
                self.data_manager.stock_dict_primary, self.data_manager.stock_dict_secondary = self.data_manager.loadDatabaseOrFetch(
                    downloadOnly, self.data_manager.list_stock_codes, menuOption, indexOption
                )
                
        self.data_manager.load_count = len(self.data_manager.stock_dict_primary) if self.data_manager.stock_dict_primary is not None else 0
        
        # Run the scanning process
        if menuOption in ["X", "B", "G", "C", "F"]:
            self.scan_executor.screen_results, self.scan_executor.save_results, self.scan_executor.backtest_df = self.scan_executor.runScanners(
                menuOption, [], self.scan_executor.tasks_queue, self.scan_executor.results_queue,
                len(self.data_manager.list_stock_codes), 0, 1, self.scan_executor.consumers,
                self.scan_executor.screen_results, self.scan_executor.save_results, self.scan_executor.backtest_df, testing
            )
            
            # Process and display results
            if not downloadOnly and menuOption in ["X", "G", "C", "F"]:
                if menuOption == "G":
                    self.user_passed_args.backtestdaysago = 0  # backtestPeriod would be set appropriately
                    
                if self.scan_executor.screen_results is not None and len(self.scan_executor.screen_results) > 0:
                    self.scan_executor.screen_results, self.scan_executor.save_results = self.result_processor.labelDataForPrinting(
                        self.scan_executor.screen_results, self.scan_executor.save_results, volumeRatio, executeOption, reversalOption or respChartPattern, menuOption
                    )
                    
                # Remove unknown values if configured
                if not self.menu_manager.newlyListedOnly and not self.config_manager.showunknowntrends and self.scan_executor.screen_results is not None and len(self.scan_executor.screen_results) > 0 and not self.user_passed_args.runintradayanalysis:
                    self.scan_executor.screen_results, self.scan_executor.save_results = self.result_processor.removeUnknowns(
                        self.scan_executor.screen_results, self.scan_executor.save_results
                    )
                    
                # Handle backtest results
                if menuOption == "B":
                    if self.scan_executor.backtest_df is not None and len(self.scan_executor.backtest_df) > 0:
                        ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
                        df_xray = self.backtest_manager.prepareGroupedXRay(0, self.scan_executor.backtest_df)
                        summary_df, sorting, sortKeys = self.backtest_manager.finishBacktestDataCleanup(self.scan_executor.backtest_df, df_xray)
                        while sorting:
                            sorting = self.backtest_manager.showSortedBacktestData(self.scan_executor.backtest_df, summary_df, sortKeys)
                            
        # Finish screening process
        self.finishScreening(
            downloadOnly, testing, self.data_manager.stock_dict_primary, self.data_manager.load_count,
            testBuild, self.scan_executor.screen_results, self.scan_executor.save_results, user
        )
        
        # Reset configuration to default
        self.resetConfigToDefault()
        
        # Handle Google Sheets integration if enabled
        self.handle_google_sheets_integration()
        
        # Handle pinned menu options
        self.handle_pinned_menu_options(testing)
        
        # Handle intraday analysis if requested
        if self.user_passed_args is not None and self.user_passed_args.runintradayanalysis:
            return self.result_processor.analysisFinalResults(
                self.scan_executor.screen_results, self.scan_executor.save_results, optionalFinalOutcome_df, 
                PKScanRunner.getFormattedChoices(self.user_passed_args, self.menu_manager.selected_choice)
            )
        else:
            return self.scan_executor.screen_results, self.scan_executor.save_results

    def handle_special_menu_options(self, menu_option):
        """
        Handle special menu options that require external execution.
        
        Args:
            menu_option: Selected menu option
        """
        launcher = f'"{sys.argv[0]}"' if " " in sys.argv[0] else sys.argv[0]
        launcher = f"python3.12 {launcher}" if (launcher.endswith(".py\"") or launcher.endswith(".py")) else launcher
        
        if menu_option == "M":
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener in monitoring mode. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} --systemlaunched -a Y -m 'X'{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit monitoring mode.{colorText.END}")
            PKAnalyticsService().send_event(f"monitor_{menu_option}")
            sleep(2)
            os.system(f"{launcher} --systemlaunched -a Y -m 'X'")
        elif menu_option == "D":
            self.handle_download_menu_option(launcher)
        elif menu_option == "L":
            PKAnalyticsService().send_event(f"{menu_option}")
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener to collect logs. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} -a Y -l{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit at any time.{colorText.END}")
            sleep(2)
            os.system(f"{launcher} -a Y -l")
        elif menu_option == "F":
            PKAnalyticsService().send_event(f"{menu_option}")
            indexOption = 0
            self.menu_manager.selected_choice["0"] = "F"
            self.menu_manager.selected_choice["1"] = "0"
            executeOption = None
            
            if self.user_passed_args is not None and self.user_passed_args.options is not None and len(self.user_passed_args.options.split(":")) >= 3:
                stockOptions = self.user_passed_args.options.split(":")[2 if len(stockOptions)<=3 else 3]
                self.data_manager.list_stock_codes = stockOptions.replace(".",",").split(",")
                
            if self.data_manager.list_stock_codes is None or len(self.data_manager.list_stock_codes) == 0:
                shouldSuppress = not OutputControls().enableMultipleLineOutput
                with SuppressOutput(suppress_stderr=shouldSuppress, suppress_stdout=shouldSuppress):
                    self.data_manager.list_stock_codes = self.data_manager.fetcher.fetchStockCodes(tickerOption=0, stockCode=None)
                    
            ConsoleUtility.PKConsoleTools.clearScreen(clearAlways=True, forceTop=True)

    def handle_download_menu_option(self, launcher):
        """
        Handle the download menu option with its sub-options.
        
        Args:
            launcher: Launcher command for external execution
        """
        selectedMenu = self.menu_manager.m0.find("D")
        ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
        self.menu_manager.m1.renderForMenu(selectedMenu)
        selDownloadOption = input(colorText.FAIL + "  [+] Select option: ") or "D"
        OutputControls().printOutput(colorText.END, end="")
        
        if selDownloadOption.upper() == "D":
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener to Download daily OHLC data. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} -a Y -e -d{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit at any time.{colorText.END}")
            PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
            sleep(2)
            os.system(f"{launcher} -a Y -e -d")
        elif selDownloadOption.upper() == "I":
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener to Download intraday OHLC data. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} -a Y -e -d -i 1m{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit at any time.{colorText.END}")
            PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
            sleep(2)
            os.system(f"{launcher} -a Y -e -d -i 1m")
        elif selDownloadOption.upper() == "N":
            self.handle_nasdaq_download_option(selectedMenu, selDownloadOption)
        elif selDownloadOption.upper() == "S":
            self.handle_sector_download_option(selectedMenu, selDownloadOption)

    def handle_nasdaq_download_option(self, selectedMenu, selDownloadOption):
        """
        Handle NASDAQ download option.
        
        Args:
            selectedMenu: Selected menu object
            selDownloadOption: Selected download option
        """
        selectedMenu = self.menu_manager.m1.find(selDownloadOption.upper())
        ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
        self.menu_manager.m2.renderForMenu(selectedMenu)
        PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
        selDownloadOption = input(colorText.FAIL + "  [+] Select option: ") or "12"
        OutputControls().printOutput(colorText.END, end="")
        
        filePrefix = "Download"
        if selDownloadOption.upper() in INDICES_MAP.keys():
            filePrefix = INDICES_MAP.get(selDownloadOption.upper()).replace(" ","")
            
        filename = (
            f"PKS_Data_{filePrefix}_"
            + PKDateUtilities.currentDateTime().strftime("%d-%m-%y_%H.%M.%S")
            + ".csv"
        )
        filePath = os.path.join(Archiver.get_user_indices_dir(), filename)
        PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
        
        if selDownloadOption.upper() == "15":
            nasdaq = PKNasdaqIndexFetcher(self.config_manager)
            _, nasdaq_df = nasdaq.fetchNasdaqIndexConstituents()
            try:
                nasdaq_df.to_csv(filePath)
            except Exception as e:
                OutputControls().printOutput(f"{colorText.FAIL}We encountered an error. Please try again!{colorText.END}\n{colorText.WARN}{e}{colorText.END}")
                pass
            OutputControls().printOutput(f"{colorText.GREEN}{filePrefix} Saved at: {filePath}{colorText.END}")
            input(f"{colorText.GREEN}Press any key to continue...{colorText.END}")
        elif selDownloadOption.upper() == "M":
            PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
        else:
            fileContents = self.data_manager.fetcher.fetchFileFromHostServer(filePath=filePath, tickerOption=int(selDownloadOption), fileContents="")
            if len(fileContents) > 0:
                OutputControls().printOutput(f"{colorText.GREEN}{filePrefix} Saved at: {filePath}{colorText.END}")
            else:
                OutputControls().printOutput(f"{colorText.FAIL}We encountered an error. Please try again!{colorText.END}")
            input(f"{colorText.GREEN}Press any key to continue...{colorText.END}")

    def handle_sector_download_option(self, selectedMenu, selDownloadOption):
        """
        Handle sector download option.
        
        Args:
            selectedMenu: Selected menu object
            selDownloadOption: Selected download option
        """
        selectedMenu = self.menu_manager.m1.find(selDownloadOption.upper())
        ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
        self.menu_manager.m2.renderForMenu(selectedMenu, skip=["15"])
        selDownloadOption = input(colorText.FAIL + "  [+] Select option: ") or "12"
        OutputControls().printOutput(colorText.END, end="")
        
        filePrefix = "Download"
        if selDownloadOption.upper() in INDICES_MAP.keys():
            filePrefix = INDICES_MAP.get(selDownloadOption.upper()).replace(" ","")
            
        filename = (
            f"PKS_Data_{filePrefix}_"
            + PKDateUtilities.currentDateTime().strftime("%d-%m-%y_%H.%M.%S")
            + ".csv"
        )
        PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
        filePath = os.path.join(Archiver.get_user_reports_dir(), filename)
        
        if selDownloadOption.upper() == "M":
            return
        else:
            indexOption = int(selDownloadOption)
            if indexOption > 0 and indexOption <= 14:
                PKAnalyticsService().send_event(f"D_{selDownloadOption.upper()}")
                shouldSuppress = not OutputControls().enableMultipleLineOutput
                with SuppressOutput(suppress_stderr=shouldSuppress, suppress_stdout=shouldSuppress):
                    self.data_manager.list_stock_codes = self.data_manager.fetcher.fetchStockCodes(indexOption, stockCode=None)
                    
                OutputControls().printOutput(f"{colorText.GREEN}Please be patient. It might take a while...{colorText.END}")
                from pkscreener.classes.PKDataService import PKDataService
                dataSvc = PKDataService()
                stockDictList, leftOutStocks = dataSvc.getSymbolsAndSectorInfo(self.config_manager, stockCodes=self.data_manager.list_stock_codes)
                
                if len(stockDictList) > 0:
                    sector_df = pd.DataFrame(stockDictList)
                    sector_df.to_csv(filePath)
                    OutputControls().printOutput(f"{colorText.GREEN}Sector/Industry info for {filePrefix}, saved at: {filePath}{colorText.END}")
                else:
                    OutputControls().printOutput(f"{colorText.FAIL}We encountered an error. Please try again!{colorText.END}")
                    
                input(f"{colorText.GREEN}Press any key to continue...{colorText.END}")

    def handle_strategy_screening(self, options):
        """
        Handle strategy screening menu option.
        
        Args:
            options: Options list
            
        Returns:
            list: Strategy filter list
        """
        strategyFilter = []
        
        if len(options) >= 2:
            userOption = options[1]
            
        if self.default_answer is None:
            selectedMenu = self.menu_manager.m0.find("S")
            self.menu_manager.m1.strategyNames = PortfolioXRay.strategyNames()
            self.menu_manager.m1.renderForMenu(selectedMenu=selectedMenu)
            
            try:
                userOption = input(colorText.FAIL + "  [+] Select option: ")
                OutputControls().printOutput(colorText.END, end="")
                
                if userOption == "":
                    userOption = "37"  # NoFilter
                elif userOption == "38":
                    userOption = input(colorText.FAIL + "  [+] Enter Exact Pattern name:")
                    OutputControls().printOutput(colorText.END, end="")
                    
                    if userOption == "":
                        userOption = "37"  # NoFilter
                    else:
                        strategyFilter.append(f"[P]{userOption}")
                        userOption = "38"
            except EOFError:
                userOption = "37"  # NoFilter
                pass
            except Exception as e:
                default_logger().debug(e, exc_info=True)
                pass
                
        userOption = userOption.upper()
        
        if userOption == "M":
            ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
            return None
        elif userOption == "Z":
            self.menu_manager.handleExitRequest(userOption)
            return None
        elif userOption == "S":
            OutputControls().printOutput(
                colorText.GREEN
                + "  [+] Collecting all metrics for summarising..."
                + colorText.END
            )
            
            # Enable showing/saving past strategy data
            savedValue = self.config_manager.showPastStrategyData
            self.config_manager.showPastStrategyData = True
            df_all = PortfolioXRay.summariseAllStrategies()
            
            if df_all is not None and len(df_all) > 0:
                OutputControls().printOutput(
                    colorText.miniTabulator().tabulate(
                        df_all,
                        headers="keys",
                        tablefmt=colorText.No_Pad_GridFormat,
                        showindex=False,
                        maxcolwidths=Utility.tools.getMaxColumnWidths(df_all)
                    ).encode("utf-8").decode(STD_ENCODING)
                )
                self.backtest_manager.showBacktestResults(df_all, sortKey="Scanner", optionalName="InsightsSummary")
            else:
                OutputControls().printOutput("[!] Nothing to show here yet. Check back later.")
                
            # Reinstate whatever was the earlier saved value
            self.config_manager.showPastStrategyData = savedValue
            
            if self.default_answer is None:
                OutputControls().takeUserInput("Press <Enter> to continue...")
                
            return None
        else:
            userOptions = userOption.split(",")
            for usrOption in userOptions:
                strategyFilter.append(self.menu_manager.m1.find(usrOption).menuText.strip())
                
        return strategyFilter

    def finishScreening(self, downloadOnly, testing, stockDictPrimary, loadCount, 
                       testBuild, screenResults, saveResults, user=None):
        """
        Finalize the screening process by saving data and notifying results.
        
        Args:
            downloadOnly (bool): Whether only downloading data without screening
            testing (bool): Whether running in test mode
            stockDictPrimary (dict): Primary stock data dictionary
            loadCount (int): Number of stocks loaded
            testBuild (bool): Whether running test build
            screenResults: Screen results data
            saveResults: Results to save
            user: User identifier for notifications
        """
        if "RUNNER" not in os.environ.keys() or downloadOnly:
            self.data_manager.saveDownloadedData(downloadOnly, testing, stockDictPrimary, loadCount)
        
        if not testBuild and not downloadOnly and not testing and \
           ((self.user_passed_args is not None and "|" not in self.user_passed_args.options) or self.user_passed_args is None):
            self.result_processor.saveNotifyResultsFile(
                screenResults, saveResults, self.default_answer, 
                self.menu_manager.menu_choice_hierarchy, user=user
            )
        
        if ("RUNNER" in os.environ.keys() and not downloadOnly) or self.user_passed_args.log:
            self.telegram_notifier.sendMessageToTelegramChannel(mediagroup=True, user=self.user_passed_args.user)

    def resetConfigToDefault(self, force=False):
        """
        Reset configuration to default values.
        
        Args:
            force (bool): Whether to force reset
        """
        if self.user_passed_args is not None and self.user_passed_args.monitor is None:
            if "PKDevTools_Default_Log_Level" in os.environ.keys():
                if self.user_passed_args is None or (self.user_passed_args is not None and self.user_passed_args.options is not None and "|" not in self.user_passed_args.options and not self.user_passed_args.runintradayanalysis and self.user_passed_args.pipedtitle is None):
                    del os.environ['PKDevTools_Default_Log_Level']
                    
        self.config_manager.logsEnabled = False
        
        if force:
            self.config_manager.logsEnabled = False
            
        self.config_manager.setConfig(ConfigManager.parser, default=True, showFileCreatedText=False)

    def handle_google_sheets_integration(self):
        """Handle Google Sheets integration if enabled."""
        try:
            creds = None
            # Write into sheet only if it's the regular scan alert trigger in the morning and evening
            if 'ALERT_TRIGGER' in os.environ.keys() and os.environ["ALERT_TRIGGER"] == 'Y':
                if "GSHEET_SERVICE_ACCOUNT_DEV" in os.environ.keys() and (self.user_passed_args.backtestdaysago is None):
                    begin = time.time()
                    creds = os.environ.get("GSHEET_SERVICE_ACCOUNT_DEV")
                    OutputControls().printOutput(f"{colorText.GREEN}  [+] Saving data to Google Spreadsheets now...{colorText.END}")
                    gClient = PKSpreadsheets(credentialDictStr=creds)
                    runOption = PKScanRunner.getFormattedChoices(self.user_passed_args, self.menu_manager.selected_choice)
                    df = self.scan_executor.saveResults.copy()
                    df["LastTradeDate"], df["LastTradeTime"] = self.data_manager.getLatestTradeDateTime(self.data_manager.stock_dict_primary)
                    gClient.df_to_sheet(df=df, sheetName=runOption)
                    OutputControls().printOutput(f"{colorText.GREEN} => Done in {round(time.time()-begin,2)}s{colorText.END}")
        except:
            pass

    def handle_pinned_menu_options(self,testing):
        """Handle pinned menu options for monitoring and quick access."""
        if ("RUNNER" not in os.environ.keys() and 
            not testing and 
            (self.user_passed_args is None or 
                (self.user_passed_args is not None and 
                    (self.user_passed_args.user is None or 
                        str(self.user_passed_args.user) == self.telegram_notifier.DEV_CHANNEL_ID) and 
                    (self.user_passed_args.answerdefault is None or self.user_passed_args.systemlaunched))) and
                    not self.user_passed_args.testbuild and 
                    self.user_passed_args.monitor is None):
            
            prevOutput_results = self.scan_executor.saveResults.index if (self.scan_executor.saveResults is not None and not self.scan_executor.saveResults.empty) else []
            isNotPiped = (("|" not in self.user_passed_args.options) if (self.user_passed_args is not None and self.user_passed_args.options is not None) else True)
            hasFoundStocks = len(prevOutput_results) > 0 and isNotPiped
            
            if hasFoundStocks or (self.config_manager.showPinnedMenuEvenForNoResult and isNotPiped):
                monitorOption = self.user_passed_args.systemlaunched if (self.user_passed_args is not None and isinstance(self.user_passed_args.systemlaunched, str) and self.user_passed_args.systemlaunched is not None) else (self.user_passed_args.options if (self.user_passed_args is not None and self.user_passed_args.options is not None) else "")
                
                if len(monitorOption) == 0:
                    for choice in self.menu_manager.selected_choice.keys():
                        monitorOption = (f"{monitorOption}:" if len(monitorOption) > 0 else '') + f"{self.menu_manager.selected_choice[choice]}"
                        
                self.menu_manager.m0.renderPinnedMenu(substitutes=[monitorOption, len(prevOutput_results), monitorOption, monitorOption, monitorOption], skip=(["1","2","4","5"] if self.menu_manager.selected_choice["0"] in ["F"] else []))
                pinOption = input(colorText.FAIL + "  [+] Select option: ") or 'M'
                OutputControls().printOutput(colorText.END, end="")
                
                self.menu_manager.ensureMenusLoaded(self.menu_manager.selected_choice["0"], self.menu_manager.selected_choice["1"], self.menu_manager.selected_choice["2"])
                
                if not PKPremiumHandler.hasPremium(self.menu_manager.m0.find(str(pinOption).upper())):
                    PKAnalyticsService().send_event(f"non_premium_user_pin_{self.menu_manager.selected_choice['0']}_{self.menu_manager.selected_choice['1']}_{self.menu_manager.selected_choice['2']}_{pinOption}")
                    PKAnalyticsService().send_event("app_exit")
                    sys.exit(0)
                    
                if pinOption in ["1", "2"]:
                    self.handle_pinned_monitoring_options(pinOption, monitorOption, prevOutput_results)
                elif pinOption in ["3"]:
                    self.handle_time_window_navigation()
                elif pinOption in ["4"]:
                    self.handle_pinned_option_saving(monitorOption)
                elif pinOption in ["5"]:
                    self.handle_restart_with_previous_results(prevOutput_results)
                    
                self.result_processor.show_saved_diff_results = False

    def handle_pinned_monitoring_options(self, pinOption, monitorOption, prevOutput_results):
        """
        Handle pinned monitoring options.
        
        Args:
            pinOption: Pinned option selected
            monitorOption: Monitor option string
            prevOutput_results: Previous results
        """
        if pinOption in ["2"]:
            monitorOption = "X:0:0"
            prevOutput_results = ",".join(prevOutput_results)
            monitorOption = f"{monitorOption}:{prevOutput_results}"
            
        launcher = f'"{sys.argv[0]}"' if " " in sys.argv[0] else sys.argv[0]
        launcher = f"python3.12 {launcher}" if (launcher.endswith(".py\"") or launcher.endswith(".py")) else launcher
        monitorOption = f'"{monitorOption}"'
        scannerOptionQuoted = monitorOption.replace("'", '"')
        
        OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener with pinned scan option. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} --systemlaunched -a Y -m {scannerOptionQuoted}{colorText.END}")
        sleep(2)
        os.system(f"{launcher} --systemlaunched -a Y -m {scannerOptionQuoted}")

    def handle_time_window_navigation(self):
        """Handle time window navigation for intraday analysis."""
        from pkscreener.classes.keys import getKeyBoardArrowInput
        
        message = f"\n  [+] {colorText.FAIL}Please use {colorText.END}{colorText.GREEN}Left / Right arrow keys{colorText.END} to slide through the {colorText.WARN}time-window by every 1 minute.{colorText.END}\n  [+] Use {colorText.GREEN}Up / Down arrow keys{colorText.END} to jump {colorText.GREEN}forward / backwards{colorText.END} by {colorText.WARN}{self.config_manager.duration}{colorText.END}\n  [+] {colorText.FAIL}Press any oher key to cancel.{colorText.END}"
        currentTime = PKDateUtilities.currentDateTime()
        requestTime = PKDateUtilities.currentDateTime()
        
        OutputControls().printOutput(message)
        direction = getKeyBoardArrowInput(message=None)
        numRequestsInASecond = 0
        
        while (direction is not None and direction not in ["RETURN", "CANCEL"]):
            requestTimeDiff = PKDateUtilities.currentDateTime() - requestTime
            
            if requestTimeDiff.total_seconds() <= 0.4:
                numRequestsInASecond += 1
            else:
                numRequestsInASecond = 0
                
            if numRequestsInASecond >= 10:
                numRequestsInASecond = 0
                fastMultiplier = 60
            else:
                fastMultiplier = 1
                
            candleFrequency = self.config_manager.candleDurationFrequency
            candleDuration = self.config_manager.candleDurationInt
            multiplier = fastMultiplier * (60 if candleFrequency == "h" else (24*60 if candleFrequency == "d" else (24*60*5 if candleFrequency == "wk" else (24*60*5*20 if candleFrequency == "mo" else 1))))
            
            if direction in ["LEFT", "DOWN"]:
                prevTime = currentTime - timedelta(minutes=(candleDuration*multiplier if direction == "DOWN" else 1*fastMultiplier))
                minPastDate = PKDateUtilities.currentDateTime() - timedelta(days=364)
                
                if prevTime <= minPastDate:
                    prevTime = minPastDate
                    
                currentTime = prevTime
                prevTime_comps = prevTime.strftime("%Y-%m-%d %H:%M:%S").split(" ")
                dateComp = prevTime_comps[0]
                timeComp = prevTime_comps[1].split(":")
                prevTime = f"{colorText.FAIL}{dateComp}{colorText.END} {prevTime_comps[1]}" if direction == "DOWN" else f"{dateComp} {colorText.FAIL}{timeComp[0]}:{timeComp[1]}{colorText.END}:{timeComp[2]}"
                
                OutputControls().moveCursorUpLines(lines=5)
                OutputControls().printOutput(message)
                OutputControls().printOutput(f"  [+] {colorText.WARN}Go back to: {colorText.END}{colorText.GREEN}{prevTime}{colorText.END}{colorText.WARN} ? Press <Enter> to confirm.{colorText.END}")
            elif direction in ["RIGHT", "UP"]:
                prevTime = currentTime + timedelta(minutes=(candleDuration*multiplier if direction == "UP" else 1*fastMultiplier))
                
                if prevTime > PKDateUtilities.currentDateTime():
                    prevTime = PKDateUtilities.currentDateTime()
                    
                currentTime = prevTime
                prevTime_comps = prevTime.strftime("%Y-%m-%d %H:%M:%S").split(" ")
                dateComp = prevTime_comps[0]
                timeComp = prevTime_comps[1].split(":")
                prevTime = f"{colorText.FAIL}{dateComp}{colorText.END} {prevTime_comps[1]}" if direction == "UP" else f"{dateComp} {colorText.FAIL}{timeComp[0]}:{timeComp[1]}{colorText.END}:{timeComp[2]}"
                
                OutputControls().moveCursorUpLines(lines=5)
                OutputControls().printOutput(message)
                OutputControls().printOutput(f"  [+] {colorText.WARN}Go forward to: {colorText.END}{colorText.GREEN}{prevTime}{colorText.END}{colorText.WARN} ? Press <Enter> to confirm.{colorText.END}")
                
            requestTime = PKDateUtilities.currentDateTime()
            direction = getKeyBoardArrowInput(message=None)
            
        if direction is not None and direction == "RETURN":
            # We need to take the data until "currentTime" from intraday data
            if self.user_passed_args is not None and self.user_passed_args.progressstatus is not None:
                runOptionName = self.user_passed_args.progressstatus.split("=>")[0].split("  [+] ")[1].strip()
                
                if runOptionName.startswith("P"):
                    self.user_passed_args.options = runOptionName.replace("_", ":")
                    
            self.user_passed_args.stocklist = ','.join(self.scan_executor.screen_results.index)
            tradingDaysInThePast = PKDateUtilities.trading_days_between(currentTime, PKDateUtilities.tradingDate())
            
            if tradingDaysInThePast > 0:
                self.user_passed_args.backtestdaysago = tradingDaysInThePast
            elif tradingDaysInThePast < 0:
                self.user_passed_args.backtestdaysago = None
            elif tradingDaysInThePast == 0:
                self.user_passed_args.slicewindow = f"'{currentTime}'"
                
            ConsoleUtility.PKConsoleTools.clearScreen(clearAlways=True, forceTop=True)
            OutputControls().printOutput(f"{colorText.WARN}Launching into the selected time-window!{colorText.END}{colorText.GREEN} Brace yourself for the time-travel!{colorText.END}")
            sleep(5)
            
            # Restart the main method with new parameters
            return self.main(userArgs=self.user_passed_args, optionalFinalOutcome_df=None)

    def handle_pinned_option_saving(self, monitorOption):
        """
        Handle saving pinned options for monitoring.
        
        Args:
            monitorOption: Monitor option to save
        """
        prevMonitorOption = f"{self.config_manager.myMonitorOptions}~" if len(self.config_manager.myMonitorOptions) > 0 else ""
        self.config_manager.myMonitorOptions = f"{prevMonitorOption}{monitorOption}"
        self.config_manager.setConfig(ConfigManager.parser, default=True, showFileCreatedText=False)
        
        OutputControls().printOutput(f"[+] {colorText.GREEN} Your monitoring options have been updated:{colorText.END}\n     {colorText.WARN}{self.config_manager.myMonitorOptions}{colorText.END}\n[+] {colorText.GREEN}You can run it using the option menu {colorText.END}{colorText.FAIL}-m{colorText.END} {colorText.GREEN}or using the main launch option {colorText.END}{colorText.FAIL}M >{colorText.END}")
        sleep(4)

    def handle_restart_with_previous_results(self, prevOutput_results):
        """
        Handle restarting with previous results.
        
        Args:
            prevOutput_results: Previous results to use
        """
        if len(self.scan_executor.save_results) > 0:
            self.data_manager.last_scan_output_stock_codes = list(self.scan_executor.save_results.index)
            self.user_passed_args.options = None
            menuOption = None
            executeOption = 0
            
            # Restart the main method with previous results
            return self.main(userArgs=self.user_passed_args, optionalFinalOutcome_df=None)

    def startMarketMonitor(self, mp_dict, keyboardevent):
        """
        Start monitoring the market status in a separate process.
        
        Args:
            mp_dict: Multiprocessing dictionary
            keyboardevent: Keyboard interrupt event
        """
        if not 'pytest' in sys.modules:
            from PKDevTools.classes.NSEMarketStatus import NSEMarketStatus
            NSEMarketStatus(mp_dict, keyboardevent).startMarketMonitor()

# Global functions for backward compatibility
def startMarketMonitor(mp_dict, keyboardevent):
    """Start market monitor - maintained for backward compatibility."""
    screener = PKScreenerMain()
    screener.startMarketMonitor(mp_dict, keyboardevent)


if __name__ == "__main__":
    screener = PKScreenerMain()
    screener.main()


"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
import sys
import pandas as pd
import numpy as np
from time import sleep
from PKDevTools.classes.Singleton import SingletonType, SingletonMixin
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes import Archiver
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKDevTools.classes.log import default_logger

# This code defines a MarketMonitor class that manages the monitoring of stock market data. 
# It utilizes the Singleton design pattern to ensure only one instance of the class exists. 
# The class initializes with a list of monitors and various configuration parameters, 
# and provides methods for updating and displaying stock data, handling alerts, 
# and formatting output for a console or Telegram integration. 
# Key functionalities include refreshing the monitor data, saving results, 
# and managing display options based on user-defined settings.

class MarketMonitor(SingletonMixin, metaclass=SingletonType):
    def __init__(self,monitors=[], maxNumResultsPerRow=3,maxNumColsInEachResult=6,maxNumRowsInEachResult=10,maxNumResultRowsInMonitor=2,pinnedIntervalWaitSeconds=30,alertOptions=[]):
        super(MarketMonitor, self).__init__()
        if monitors is not None and len(monitors) > 0:
            
            self.monitors = monitors[:maxNumResultRowsInMonitor*maxNumResultsPerRow]
            self.monitorIndex = 0
            self.monitorPositions = {}
            self.monitorResultStocks = {}
            self.alertOptions = alertOptions
            self.hiddenColumns = ""
            self.alertStocks = []
            self.alertedStocks = {}
            self.pinnedIntervalWaitSeconds = pinnedIntervalWaitSeconds
            # self.monitorNames = {}
            # We are going to present the dataframes in a 3x3 matrix with limited set of columns
            rowIndex = 0
            colIndex = 0
            self.maxNumResultRowsInMonitor = maxNumResultRowsInMonitor
            self.maxNumRowsInEachResult = maxNumRowsInEachResult
            self.maxNumColsInEachResult = maxNumColsInEachResult
            self.maxNumResultsPerRow = maxNumResultsPerRow
            maxColIndex = self.maxNumColsInEachResult * self.maxNumResultsPerRow - 1
            self.lines = 0
            monIndex = 0
            for monitorKey in self.monitors:
                self.alertedStocks[str(monIndex)] = []
                monIndex += 1
                self.monitorPositions[monitorKey] = [rowIndex,colIndex]
                # self.monitorNames[monitorKey] = ""
                colIndex += self.maxNumColsInEachResult
                if colIndex > maxColIndex:
                    colIndex = 0
                    rowIndex += self.maxNumRowsInEachResult
            columns = []
            colNameIndex = 0
            maxColIndex = min(maxColIndex,len(self.monitorPositions)*self.maxNumColsInEachResult -1)
            while colNameIndex <= maxColIndex:
                columns.append(f"A{colNameIndex +1}")
                colNameIndex += 1
            self.monitor_df = pd.DataFrame(columns=columns)
            self.isPinnedSingleMonitorMode = len(self.monitorPositions.keys()) == 1

    def currentMonitorOption(self):
        try:
            option = None
            maxIndex = len(self.monitors) -1
            option = str(self.monitors[self.monitorIndex:self.monitorIndex+1][0])
            self.monitorIndex += 1
            if self.monitorIndex > maxIndex:
                self.monitorIndex = 0
        except: # pragma: no cover
            pass
        return option

    def saveMonitorResultStocks(self, results_df):
        try:
            if len(self.alertedStocks.keys()) < self.monitorIndex+1: # for pytests
                self.alertedStocks[str(self.monitorIndex)] = []
            self.alertStocks = []
            lastSavedResults = self.monitorResultStocks[str(self.monitorIndex)]
            lastSavedResults = lastSavedResults.split(",")
        except:
            lastSavedResults = []
            pass
        prevOutput_results = ""
        if results_df is None or results_df.empty:
            prevOutput_results = "NONE"
        else:
            prevOutput_results = results_df[~results_df.index.duplicated(keep='first')]
            prevOutput_results = prevOutput_results.index
            prevOutput_results = ",".join(prevOutput_results)
        if len(self.monitorResultStocks.keys()) > self.monitorIndex and str(self.monitorIndex) in self.monitorResultStocks.keys() and len(self.monitorResultStocks[str(self.monitorIndex)]) > 0 and prevOutput_results == "NONE":
            prevOutput_results = self.monitorResultStocks[str(self.monitorIndex)]
        addedStocks = list(set(prevOutput_results.split(',')) - set(lastSavedResults))  # Elements in new df but not in the saved df
        if len(self.alertStocks) != len(addedStocks) and len(addedStocks) > 0:
            self.alertStocks = addedStocks
        diffAlerts = list(set(self.alertStocks) - set(self.alertedStocks[str(self.monitorIndex)]))
        if len(diffAlerts) > 0:
            self.alertStocks = diffAlerts
            self.alertedStocks[str(self.monitorIndex)].extend(diffAlerts)
        else:
            self.alertStocks = []
        if len(addedStocks) > 0:
            self.monitorResultStocks[str(self.monitorIndex)] = prevOutput_results

    def refresh(self, screen_df:pd.DataFrame=None, screenOptions=None, chosenMenu="", dbTimestamp="", telegram=False):
        highlightRows = []
        highlightCols = []
        telegram_df = None
        if screen_df is None or screen_df.empty or screenOptions is None:
            return
        screen_monitor_df = screen_df.copy()
        monitorPosition = self.monitorPositions.get(screenOptions)

        from pkscreener.classes import Utility, ImageUtility

        if self.isPinnedSingleMonitorMode:
            screen_monitor_df = screen_monitor_df[screen_monitor_df.columns[:14]]
            self.monitor_df = screen_monitor_df
            if "RUNNER" in os.environ.keys():
                self.monitor_df.reset_index(inplace=True)
                with pd.option_context('mode.chained_assignment', None):
                    self.monitor_df = self.monitor_df[["Stock", "LTP", "%Chng","52Wk-H","RSI/i" if "RSI/i" in self.monitor_df.columns else "RSI","volume"]]
                    # Import Utility here since Utility has dependency on PKScheduler which in turn has dependency on 
                    # multiprocessing, which behaves erratically if imported at the top.
                    self.monitor_df.loc[:, "%Chng"] = self.monitor_df.loc[:, "%Chng"].astype(str).apply(
                                lambda x: ImageUtility.PKImageTools.roundOff(str(x).split("% (")[0] + colorText.END,0)
                            )
                    self.monitor_df.loc[:, "52Wk-H"] = self.monitor_df.loc[:, "52Wk-H"].astype(str).apply(
                        lambda x: ImageUtility.PKImageTools.roundOff(x,0)
                    )
                    self.monitor_df.loc[:, "volume"] = self.monitor_df.loc[:, "volume"].astype(str).apply(
                        lambda x: ImageUtility.PKImageTools.roundOff(x,0)
                    )
                    self.monitor_df.rename(columns={"%Chng": "Ch%","volume":"Vol","52Wk-H":"52WkH", "RSI":"RSI/i"}, inplace=True)
                    telegram_df = self.updateDataFrameForTelegramMode(telegram or "RUNNER" in os.environ.keys(), self.monitor_df)
                    self.monitor_df.set_index("Stock",inplace=True)
            
        if not self.isPinnedSingleMonitorMode:
            screen_monitor_df.reset_index(inplace=True)
            with pd.option_context('mode.chained_assignment', None):
                screen_monitor_df = screen_monitor_df[["Stock", "LTP", "%Chng","52Wk-H","RSI/i" if "RSI/i" in screen_monitor_df.columns else "RSI","volume"]].head(self.maxNumRowsInEachResult-1)
                # Import Utility here since Utility has dependency on PKScheduler which in turn has dependency on 
                # multiprocessing, which behaves erratically if imported at the top.
                screen_monitor_df.loc[:, "%Chng"] = screen_monitor_df.loc[:, "%Chng"].astype(str).apply(
                            lambda x: ImageUtility.PKImageTools.roundOff(str(x).split("% (")[0] + colorText.END,0)
                        )
                screen_monitor_df.loc[:, "52Wk-H"] = screen_monitor_df.loc[:, "52Wk-H"].astype(str).apply(
                    lambda x: ImageUtility.PKImageTools.roundOff(x,0)
                )
                screen_monitor_df.loc[:, "volume"] = screen_monitor_df.loc[:, "volume"].astype(str).apply(
                    lambda x: ImageUtility.PKImageTools.roundOff(x,0)
                )
                screen_monitor_df.rename(columns={"%Chng": "Ch%","volume":"Vol","52Wk-H":"52WkH", "RSI":"RSI/i"}, inplace=True)
            telegram_df = self.updateDataFrameForTelegramMode(telegram or "RUNNER" in os.environ.keys(), screen_monitor_df)
        
        if monitorPosition is not None:
            startRowIndex, startColIndex = monitorPosition
            if not self.monitor_df.empty:
                OutputControls().moveCursorUpLines(self.lines)
            if not self.isPinnedSingleMonitorMode:
                firstColIndex = startColIndex
                rowIndex = 0
                colIndex = 0
                highlightRows = [startRowIndex]
                highlightCols = []
                while rowIndex <= len(screen_monitor_df):
                    for col in screen_monitor_df.columns:
                        if rowIndex == 0:
                            # Column names to be repeated for each refresh in respective headers
                            cleanedScreenOptions = screenOptions.replace(":D","")
                            widgetHeader = self.getScanOptionName(cleanedScreenOptions)
                            if len(widgetHeader) <= 0:
                                if cleanedScreenOptions.startswith("|"):
                                    cleanedScreenOptions = cleanedScreenOptions.replace("|","")
                                    pipedFrom = ""
                                    if cleanedScreenOptions.startswith("{"):
                                        pipedFrom = cleanedScreenOptions.split("}")[0] + "}:"
                                    cleanedScreenOptions = pipedFrom + ":".join(cleanedScreenOptions.split(":")[2:])
                                    cleanedScreenOptions = cleanedScreenOptions.replace(">X:0:","")
                                widgetHeader = ":".join(cleanedScreenOptions.split(":")[:4])
                                if "i " in screenOptions:
                                    widgetHeader = f'{":".join(widgetHeader.split(":")[:3])}:i:{cleanedScreenOptions.split("i ")[-1]}'
                            self.monitor_df.loc[startRowIndex,[f"A{startColIndex+1}"]] = colorText.HEAD+(widgetHeader if startColIndex==firstColIndex else col)+colorText.END
                            highlightCols.append(startColIndex)
                        else:
                            if colIndex == 0:
                                stockNameDecorated = screen_monitor_df.iloc[rowIndex-1,colIndex]
                                stockName = ImageUtility.PKImageTools.stockNameFromDecoratedName(stockNameDecorated)
                                stockName = (f"{colorText.BOLD}{colorText.WHITE_FG_BRED_BG}{stockNameDecorated}{colorText.END}") if stockName in self.alertStocks else stockNameDecorated
                                self.monitor_df.loc[startRowIndex, [f"A{startColIndex+1}"]] = stockName
                            else:
                                self.monitor_df.loc[startRowIndex, [f"A{startColIndex+1}"]] = screen_monitor_df.iloc[rowIndex-1,colIndex]
                            colIndex += 1
                        startColIndex += 1
                    _, startColIndex= monitorPosition
                    rowIndex += 1
                    colIndex = 0
                    highlightRows.append(startRowIndex+1)
                    startRowIndex += 1
            else:
                stocks = list(self.monitor_df.index)
                updatedStocks = []
                for stock in stocks:
                    stockName = ImageUtility.PKImageTools.stockNameFromDecoratedName(stock)
                    stockName = (f"{colorText.BOLD}{colorText.WHITE_FG_BRED_BG}{stock}{colorText.END}") if stockName in self.alertStocks else stock
                    updatedStocks.append(stockName)
                self.monitor_df.reset_index(inplace=True)
                with pd.option_context('mode.chained_assignment', None):
                    self.monitor_df["Stock"] = updatedStocks
                self.monitor_df.set_index("Stock",inplace=True)

        self.monitor_df = self.monitor_df.replace(np.nan, "-", regex=True)
        latestScanMenuOption = f"  [+] {dbTimestamp} (Dashboard) > " + f"{chosenMenu[:190]} [{screenOptions}]"
        OutputControls().printOutput(
            colorText.FAIL
            + latestScanMenuOption[:200]
            + colorText.END
            , enableMultipleLineOutput=True
        )
        tabulated_results = colorText.miniTabulator().tabulate(
            self.monitor_df, tablefmt=colorText.No_Pad_GridFormat,
            headers="keys" if self.isPinnedSingleMonitorMode else (),
            highlightCharacter=colorText.HEAD+"="+colorText.END,
            showindex=self.isPinnedSingleMonitorMode,
            highlightedRows=highlightRows,
            highlightedColumns=highlightCols,
            maxcolwidths=Utility.tools.getMaxColumnWidths(self.monitor_df)
        )
        console_results = ""
        if self.isPinnedSingleMonitorMode:
            copyScreenResults = self.monitor_df.copy()
            hiddenColumns = self.hiddenColumns.split(",")
            for col in copyScreenResults.columns:
                if col in hiddenColumns:
                    copyScreenResults.drop(col, axis=1, inplace=True, errors="ignore")
            try:
                console_results = colorText.miniTabulator().tabulate(
                        copyScreenResults, headers="keys", tablefmt=colorText.No_Pad_GridFormat,
                        maxcolwidths=Utility.tools.getMaxColumnWidths(copyScreenResults)
                    )
            except: # pragma: no cover
                console_results = tabulated_results
        numRecords = len(tabulated_results.splitlines())
        self.lines = numRecords #+ 1 #(1 if len(self.monitorResultStocks.keys()) <= self.maxNumColsInEachResult else 0) # 1 for the progress bar at the bottom and 1 for the chosenMenu option
        OutputControls().printOutput(tabulated_results if not self.isPinnedSingleMonitorMode else console_results, enableMultipleLineOutput=True)
        
        if not telegram and ((screenOptions in self.alertOptions and numRecords > 1) or len(self.alertStocks) > 0): # Alert conditions met? Sound alert!
            Utility.tools.alertSound(beeps=5)
        if not self.isPinnedSingleMonitorMode:
            if telegram:
                self.updateIfRunningInTelegramBotMode(screenOptions, chosenMenu, dbTimestamp, telegram, telegram_df)
        else:
            if ((screenOptions in self.alertOptions and numRecords > 3) or len(self.alertStocks) > 0): # Alert conditions met? Sound alert!
                # numRecords is actually new lines. Top 3 lines are only headers
                if telegram_df is not None:
                    telegram_df.reset_index(inplace=True)
                    notify_df = telegram_df[telegram_df["Stock"].isin(self.alertStocks)]
                    notify_df = notify_df[["Stock","LTP","Ch%","Vol"]].head(50)
                    if len(notify_df) > 0:
                        notify_output = self.updateIfRunningInTelegramBotMode(screenOptions, chosenMenu, dbTimestamp, False, notify_df,maxcolwidths=None)
                        if len(notify_output) > 0:
                            from PKDevTools.classes.pubsub.publisher import PKUserService
                            from PKDevTools.classes.pubsub.subscriber import notification_service
                            PKUserService().notify_user(scannerID=self.getScanOptionName(screenOptions),notification=notify_output)
                    # notify_df = self.monitor_df.reindex(self.alertStocks)  # Includes missing stocks, if any. Returns NaN for such cases
                Utility.tools.alertSound(beeps=5)
            sleep(self.pinnedIntervalWaitSeconds)

    def updateDataFrameForTelegramMode(self, telegram, screen_monitor_df):
        telegram_df = None
        if telegram:
            telegram_df = screen_monitor_df[["Stock", "LTP", "Ch%", "Vol"]]
            try:
                telegram_df.loc[:, "Stock"] = telegram_df.loc[:, "Stock"].apply(
                    lambda x: x.split('\x1b')[3].replace('\\','') if 'http' in x else x
                )
                cols = ["LTP", "Ch%", "Vol"]
                for col in cols:
                    telegram_df.loc[:, col] = telegram_df.loc[:, col].apply(
                        lambda x: x.replace(colorText.FAIL,"").replace(colorText.GREEN,"").replace(colorText.WARN,"").replace(colorText.BOLD,"").replace(colorText.END,"")
                    )
                telegram_df.loc[:, "LTP"] = telegram_df.loc[:, "LTP"].apply(
                    lambda x: str(int(round(float(x),0)))
                )
                telegram_df.loc[:, "Ch%"] = telegram_df.loc[:, "Ch%"].apply(
                    lambda x: f'{int(round(float(x.replace("%","")),0))}%'
                )
                telegram_df.loc[:, "Vol"] = telegram_df.loc[:, "Vol"].apply(
                    lambda x: f'{int(round(float(x.replace("x","")),0))}x'
                )
                with SuppressOutput(suppress_stderr=True, suppress_stdout=True):
                    for col in telegram_df.columns:
                        telegram_df[col] = telegram_df[col].astype(str)
            except: # pragma: no cover
                pass
        return telegram_df

    def updateIfRunningInTelegramBotMode(self, screenOptions, chosenMenu, dbTimestamp, telegram, telegram_df,maxcolwidths=[None,None,4,3]):
        result_output = ""
        telegram_df_tabulated = ""
        if telegram_df is not None:
            STD_ENCODING=sys.stdout.encoding if sys.stdout is not None else 'utf-8'
            try:
                telegram_df_tabulated = colorText.miniTabulator().tabulate(
                                telegram_df,
                                headers="keys",
                                tablefmt=colorText.No_Pad_GridFormat,
                                showindex=False,
                                maxcolwidths=maxcolwidths if maxcolwidths is not None else None
                            ).encode("utf-8").decode(STD_ENCODING).replace("-K-----S-----C-----R","-K-----S----C---R").replace("%  ","% ").replace("=K=====S=====C=====R","=K=====S====C===R").replace("Vol  |","Vol|").replace("x  ","x")
            except Exception as e:
                default_logger().debug(e,exc_info=True)
                pass
            telegram_df_tabulated = telegram_df_tabulated.replace("-E-----N-----E-----R","-E-----N----E---R").replace("=E=====N=====E=====R","=E=====N====E===R")
            choiceSegments = chosenMenu.split(">")
            optionName = self.getScanOptionName(screenOptions)
            chosenMenu = f"{choiceSegments[-2]}>{choiceSegments[-1]}" if (len(choiceSegments)>=4 or len(choiceSegments[-1]) <= 10) else f"{choiceSegments[-1]}"
            result_output = f"Latest data as of {dbTimestamp}\n<b>[{optionName}] {chosenMenu}</b> [{screenOptions}]\n<pre>{telegram_df_tabulated}</pre>"
            try:
                if telegram:
                    filePath = os.path.join(Archiver.get_user_data_dir(), f"monitor_outputs_{self.monitorIndex}.txt")
                    f = open(filePath, "w")
                    f.write(result_output)
                    f.close()
            except: # pragma: no cover
                pass
        return result_output

    def getScanOptionName(self, screenOptions):
        from pkscreener.classes.MenuOptions import PREDEFINED_SCAN_MENU_VALUES
        if screenOptions is None:
            return ""
        baseIndex = 12
        baseIndices = str(screenOptions).split(":")
        if len(baseIndices) > 1:
            baseIndex = baseIndices[1]
        choices = f"--systemlaunched -a y -e -o '{str(screenOptions).replace('C:','X:').replace('D:','')}'"
        indexNum = -1
        try:
            indexNum = PREDEFINED_SCAN_MENU_VALUES.index(choices)
        except: # pragma: no cover
            pass
        optionName = str(screenOptions).replace(':D','').replace(':','_')
        if indexNum >= 0:
            optionName = f"{('P_1_'+str(indexNum +1)+'_'+str(baseIndex)) if '>|' in choices else str(screenOptions).replace(':D','').replace(':','_')}"
        return optionName


"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import warnings
import os
import time
from PKDevTools.classes.OutputControls import OutputControls
warnings.simplefilter("ignore", UserWarning,append=True)
os.environ["PYTHONWARNINGS"]="ignore::UserWarning"

def init_pool_processes(the_lock):
    '''Initialize each process with a global variable lock.
    '''
    global lock
    lock = the_lock

import multiprocessing
from multiprocessing import Lock

# from time import sleep
from concurrent.futures import ProcessPoolExecutor
from rich.progress import Progress, BarColumn, TimeRemainingColumn, TimeElapsedColumn
from rich.console import Console
from rich.control import Control
from rich.segment import ControlType
from pkscreener.classes.PKTask import PKTask

if __name__ == '__main__':
    multiprocessing.freeze_support()

# def long_running_fn(*args, **kwargs):
#     len_of_task = random.randint(3, 20000)  # take some random length of time
#     task = args[0]
#     progress = task.progressStatusDict
#     task_id = task.taskId
#     for n in range(0, len_of_task):
#         # sleep(1)  # sleep for a bit to simulate work
#         progress[task_id] = {"progress": n + 1, "total": len_of_task}
#     # if task is not None:
#     #     progress[task_id] = task.progress_fn(task_id)
#     # else:
#     #     progress[task_id] = {"progress": 100, "total": 100}

progressUpdater=None
class PKScheduler():
    def scheduleTasks(tasksList=[], label:str=None, showProgressBars=False,submitTaskAsArgs=True, timeout=6, minAcceptableCompletionPercentage=100):
        n_workers = multiprocessing.cpu_count() - 1  # set this to the number of cores you have on your machine
        global progressUpdater
        console = Console()
        with Progress(
                "[progress.description]{task.description}",
                BarColumn(),
                "[progress.percentage]{task.percentage:>3.0f}%",
                TimeRemainingColumn(),
                TimeElapsedColumn(),
                auto_refresh = True,
                refresh_per_second=100,  # bit slower updates if we keep it to 1
                console=console
            ) as progress:
            progressUpdater = progress
            if len(tasksList) == 0:
                raise ValueError("No tasks in the tasksList!")
            for task in tasksList:
                if not isinstance(task, PKTask):
                    raise ValueError("Each task in the tasksList must be of type PKTask!")
            futures = []  # keep track of the jobs
            with multiprocessing.Manager() as manager:
                # this is the key - we share some state between our 
                # main process and our worker functions
                _progress = manager.dict()
                _results = manager.dict()
                console.control(Control(*((ControlType.CURSOR_UP,1),))) # Cursor up 1 lines f"\x1b[{param}A"
                # sys.stdout.write("\x1b[2K")  # delete the last line
                if showProgressBars:
                    overall_progress_task = progress.add_task(f"[green]{label if label is not None else 'Pending jobs progress:'}", visible=showProgressBars)

                lock = Lock()
                taskResultsUpdated = False
                with ProcessPoolExecutor(max_workers=n_workers,
                                         initializer=init_pool_processes,
                                         initargs=(lock,)) as executor:
                    for task in tasksList:  # iterate over the jobs we need to run
                        # set visible false so we don't have a lot of bars all at once:
                        task_id = progress.add_task(f"Task :{task.taskName}", visible=showProgressBars)
                        task.taskId = task_id
                        task.progressStatusDict = _progress
                        task.resultsDict = _results
                        futures.append(executor.submit(task.long_running_fn, task if submitTaskAsArgs else task.long_running_fn_args))

                    # monitor the progress:
                    start_time = time.time()
                    while (((n_finished := int(sum([future.done() for future in futures]))) < len(futures)) and ((time.time() - start_time) < timeout)):
                        if showProgressBars:
                            progress.update(
                                overall_progress_task,
                                completed=n_finished,
                                total=len(futures),
                                visible=n_finished < len(futures)
                            )
                            OutputControls().printOutput(f"{n_finished} of {len(futures)}")
                        # We've reached a state where the caller may not want to wait any further
                        if n_finished*100/len(futures) >= minAcceptableCompletionPercentage:
                            break
                        for task_id, update_data in _progress.items():
                            for task in tasksList:
                                if task.taskId == task_id:
                                    taskResultsUpdated = True
                                    task.result = task.resultsDict.get(task_id)
                            latest = update_data["progress"]
                            total = update_data["total"]
                            if showProgressBars:
                                # update the progress bar for this task:
                                progress.update(
                                    task_id,
                                    completed=latest,
                                    total=total,
                                    visible=(latest < total) and showProgressBars,
                                )
                            lock.acquire()
                            progress.refresh()
                            lock.release()
                    # sleep(0.1)
                    if showProgressBars:
                        progress.update(
                                overall_progress_task,
                                completed=1,
                                total=1,
                                visible=False
                            )
                    for task_id, update_data in _progress.items():
                        # update the progress bar for this task:
                        if showProgressBars:
                            progress.update(
                                task_id,
                                completed=1,
                                total=1,
                                visible=False,
                            )
                    if not taskResultsUpdated:
                        for task in tasksList:
                            if task.taskId == task_id:
                                task.result = task.resultsDict.get(task_id)
                    lock.acquire()
                    progress.refresh()
                    # raise any errors:
                    # for future in futures:
                    #     future.result()
                    lock.release()

# if __name__ == "__main__":
#     scheduleTasks([PKTask("Task 1",long_running_fn,),
#                 PKTask("Task 2",long_running_fn),
#                 PKTask("Task 3",long_running_fn),
#                 PKTask("Task 4",long_running_fn),
#                 PKTask("Task 5",long_running_fn),
#                 PKTask("Task 6",long_running_fn),
#                 PKTask("Task 7",long_running_fn),
#                 PKTask("Task 8",long_running_fn),
#                 PKTask("Task 9",long_running_fn),
#                 PKTask("Task 10",long_running_fn)])


"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import datetime
import pytz
import gspread_pandas as gp
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
import json
import os

class PKSpreadsheets:

    def __init__(self,credentialFilePath=None,credentialDictStr=None) -> None:
        self.gClient = None
        self.credentialFilePath = credentialFilePath
        self.credentialDictStr = credentialDictStr
        self.credentials = None
        self.authCredentials = None

    def login(self):
        if self.credentialFilePath is not None:
            with open(self.credentialFilePath) as f:
                self.credentials = json.load(f)
        elif self.credentialDictStr is not None:
            self.credentials = json.loads(self.credentialDictStr)
        else:
            if "GSHEET_SERVICE_ACCOUNT_DEV" in os.environ.keys():
                self.credentials = json.load(os.environ["GSHEET_SERVICE_ACCOUNT_DEV"])
        if self.credentials is None:
            raise ValueError("Credentials cannot be empty!")
        
        DEFAULT_SCOPES =[
            'https://www.googleapis.com/auth/spreadsheets',
            'https://www.googleapis.com/auth/drive'
        ]
        self.authCredentials = ServiceAccountCredentials.from_service_account_info(
            info=self.credentials,
            scopes=DEFAULT_SCOPES,
        )
        self.gClient = gp.Client(creds=self.authCredentials,config=self.credentials,load_dirs=False) #gspread.service_account_from_dict(self.credentials)

    def listFolders(self):
        return self.gClient._get_dirs()
    
    def getFolder(self,path="/"):
        dirs = self.listFolders()
        # 'id': '0AEHslrN6p5ejUk9PVA'
        # 'name': 'My Drive'
        # 'path': '/'
        folder = None
        if path != "/":
            for dir in dirs:
                if dir["path"] == path:
                    folder = dir
                    break
        else:
            for dir in dirs:
                if dir["path"] == "/":
                    folder = dir
        return folder
    
    def listWorkbooks(self,path="/", folderId=None):
        if folderId is None:
            folder = self.getFolder(path=path)
            if folder is not None:
                folderId = folder["id"]
        existingWorkbooks = None
        if folderId is not None:
            existingWorkbooks = self.gClient.list_spreadsheet_files_in_folder(folder_id=folderId)
        return existingWorkbooks

    def getWorkbookByName(self,workbookName=None, atPath="/", folderId=None) -> gp.Spread:
        existingWorkbooks = self.listWorkbooks(path=atPath,folderId=folderId)
        workbookOutcome = None
        for workbook in existingWorkbooks:
            if workbook["name"] == workbookName:
                workbookOutcome = self.gClient.open_by_key(workbook["id"])
                break
        return gp.Spread(workbookOutcome.id, client=self.gClient, creds=self.authCredentials, config=self.credentials) if workbookOutcome is not None else None
    
    def createFolder(self, path="/"):
        if path == "/":
            raise ValueError("Cannot create the root folder! Supply a path value, e.g., /Parent/Child/")
        return self.gClient.create_folder(path=path)

    def createWorkbook(self, workbookName=None, atFolderPath="/"):
        if self.gClient is None:
            raise Exception("You must login first using login() method!")
        if workbookName is None:
            raise ValueError("workbookName cannot be empty!")
        workbookOutcome = None
        # 'id': '0AEHslrN6p5ejUk9PVA'
        # 'name': 'My Drive'
        # 'path': '/'
        folder = self.getFolder(path=atFolderPath)
        if folder is not None:
            folderId = folder["id"]
            workbookOutcome = self.getWorkbookByName(workbookName=workbookName,atPath=atFolderPath,folderId=folderId)
        else:
            folderInfo = self.createFolder(path=atFolderPath)
            folderId = folderInfo["id"]
        if workbookOutcome is None:
            workbookOutcome = self.gClient.create(title=workbookName,folder_id=folderId)
            # 'id':'18ijLL0uGSYTeRYFb8aQNzJc4m7lCH61T2T8wtpfodSk'
            # 'title':'PKScreener'
            # 'locale':'en_US'
            # 'autoRecalc':'ON_CHANGE'
            # 'timeZone':'Etc/GMT'
            # 'defaultFormat':{'backgroundColor': {'red': 1, 'green': 1, 'blue': 1}, 'padding': {'top': 2, 'right': 3, 'bottom': 2, 'left': 3}, 'verticalAlignment': 'BOTTOM', 'wrapStrategy': 'OVERFLOW_CELL', 'textFormat': {'foregroundColor': {}, 'fontFamily': 'arial,sans,sans-serif', 'fontSize': 10, 'bold': False, 'italic': False, 'strikethrough': False, 'underline': False, 'foregroundColorStyle': {...}}, 'backgroundColorStyle': {'rgbColor': {...}}}
            # 'spreadsheetTheme':{'primaryFontFamily': 'Arial', 'themeColors': [{...}, {...}, {...}, {...}, {...}, {...}, {...}, {...}, {...}]}
            # 'name':'PKScreener'
            # 'createdTime':'2024-03-07T19:40:58.328Z'
            # 'modifiedTime':'2024-03-07T19:40:58.412Z'
            workbookOutcome.share('pkscreener.in@gmail.com', perm_type='user', role='writer', notify=False)
            permissions = workbookOutcome.list_permissions()
            for permission in permissions:
                if permission["emailAddress"] == "pkscreener.in@gmail.com" and permission["role"] == "writer":
                    workbookOutcome.transfer_ownership(permission_id=permission["id"])
                    break
        return gp.Spread(workbookOutcome.id, client=self.gClient, creds=self.authCredentials, config=self.credentials) if workbookOutcome is not None else None

    def addWorksheet(self,worksheetName=None,workbook:gp.Spread=None):
        assert workbook is not None
        assert worksheetName is not None
        workbook.open_sheet(sheet=worksheetName, create=True)
        return workbook.sheet if workbook.sheet.title == worksheetName else None
    
    def findSheet(self, worksheetName=None,workbook:gp.Spread=None):
        return workbook.find_sheet(sheet=worksheetName)
    
    def df_to_sheet(self,df=None,sheetName=None,folderName="PKScreener",workbookName="PKScreener"):
        self.login()
        valueAddColumns = ["EntryDate", "EntryTime", "ScanLabel"]
        today = datetime.datetime.now(pytz.timezone("Asia/Kolkata"))
        date = today.date().strftime("%Y-%m-%d")
        time = today.strftime("%H:%M:%S")
        for col in valueAddColumns:
            if col not in df.columns:
                df[col] = date if col == valueAddColumns[0] else (time if col == valueAddColumns[1] else sheetName)
        workbook = self.getWorkbookByName(workbookName=workbookName,atPath=folderName)
        sheet = self.addWorksheet(worksheetName=sheetName,workbook=workbook)
        currentMaxRow = sheet.row_count
        currentMaxCol = sheet.col_count
        currentMaxRow = currentMaxRow + 1 if currentMaxRow > 1 else currentMaxRow
        sheet.resize(len(df)+ currentMaxRow, max(len(df.columns),currentMaxCol))
        workbook.df_to_sheet(df=df, start=(currentMaxRow,1), freeze_headers=False, freeze_index=True,sheet=sheet)
        sheet.freeze(rows=1, cols=1)



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""

import os
import sys
import warnings
import logging
from time import sleep
warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)
import pandas as pd
# import yfinance as yf
# from yfinance import shared
# from yfinance.const import USER_AGENTS
from PKDevTools.classes.Utils import USER_AGENTS
import random
# from yfinance.version import version as yfVersion
# if yfVersion == "0.2.28":
#     from yfinance.data import TickerData as YfData
#     class YFPricesMissingError(Exception):
#         pass
#     class YFInvalidPeriodError(Exception):
#         pass
#     class YFRateLimitError(Exception):
#         pass
# else:
#     from yfinance.data import YfData
#     from yfinance.exceptions import YFPricesMissingError, YFInvalidPeriodError, YFRateLimitError
from concurrent.futures import ThreadPoolExecutor
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.Fetcher import StockDataEmptyException
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKNSETools.PKNSEStockDataFetcher import nseStockDataFetcher
from pkscreener.classes.PKTask import PKTask
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes import Archiver
# This Class Handles Fetching of Stock Data over the internet

from requests import Session
from requests_cache import CacheMixin, SQLiteCache
from requests_ratelimiter import LimiterMixin, MemoryQueueBucket
from pyrate_limiter import Duration, RequestRate, Limiter
class CachedLimiterSession(CacheMixin, LimiterMixin, Session):
   pass

# https://help.yahooinc.com/dsp-api/docs/rate-limits
# Define multiple rate limits
TRY_FACTOR = 1
yf_limiter = Limiter(
    RequestRate(60*TRY_FACTOR, Duration.MINUTE),      # Max 60 requests per minute
    RequestRate(360*TRY_FACTOR, Duration.HOUR),       # Max 360 requests per hour
    RequestRate(8000*TRY_FACTOR, Duration.DAY)        # Max 8000 requests per day
)
# yf_session = CachedLimiterSession(
#    limiter=yf_limiter,
#    bucket_class=MemoryQueueBucket,
#    backend=SQLiteCache(db_path=os.path.join(Archiver.get_user_data_dir(),"yfinance.cache")),
# )
class screenerStockDataFetcher(nseStockDataFetcher):
    _tickersInfoDict={}
    def fetchStockDataWithArgs(self, *args):
        task = None
        if isinstance(args[0], PKTask):
            task = args[0]
            stockCode,period,duration,exchangeSuffix = task.long_running_fn_args
        else:
            stockCode,period,duration,exchangeSuffix = args[0],args[1],args[2],args[3]
        result = self.fetchStockData(stockCode,period,duration,None,0,0,0,exchangeSuffix=exchangeSuffix,printCounter=False)
        if task is not None:
            if task.taskId >= 0:
                task.progressStatusDict[task.taskId] = {'progress': 0, 'total': 1}
                task.resultsDict[task.taskId] = result
                task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
            task.result = result
        return result

    def get_stats(self,ticker):
        info = None #yf.Tickers(ticker).tickers[ticker].fast_info
        screenerStockDataFetcher._tickersInfoDict[ticker] = {"marketCap":info.market_cap if info is not None else 0}

    def fetchAdditionalTickerInfo(self,ticker_list,exchangeSuffix=".NS"):
        if not isinstance(ticker_list,list):
            raise TypeError("ticker_list must be a list")
        if len(exchangeSuffix) > 0:
            ticker_list = [(f"{x}{exchangeSuffix}" if not x.endswith(exchangeSuffix) else x) for x in ticker_list]
        screenerStockDataFetcher._tickersInfoDict = {}
        with ThreadPoolExecutor() as executor:
            executor.map(self.get_stats, ticker_list)
        return screenerStockDataFetcher._tickersInfoDict

    # Fetch stock price data from Yahoo finance
    def fetchStockData(
        self,
        stockCode,
        period,
        duration,
        proxyServer=None,
        screenResultsCounter=0,
        screenCounter=0,
        totalSymbols=0,
        printCounter=False,
        start=None, 
        end=None,
        exchangeSuffix=".NS",
        attempt = 0
    ):
        """
        if isinstance(stockCode,list):
            if len(exchangeSuffix) > 0:
                stockCode = [(f"{x}{exchangeSuffix}" if (not x.endswith(exchangeSuffix) and not x.startswith("^")) else x) for x in stockCode]
        elif isinstance(stockCode,str):
            if len(exchangeSuffix) > 0:
                stockCode = f"{stockCode}{exchangeSuffix}" if (not stockCode.endswith(exchangeSuffix) and not stockCode.startswith("^")) else stockCode
        if (period in ["1d","5d","1mo","3mo","5mo"] or duration[-1] in ["m","h"]):
            # Since this is intraday data, we'd just need to start from the last trading session
            # if start is None:
            #     start = PKDateUtilities.tradingDate().strftime("%Y-%m-%d")
            # if end is None:
            #     end = PKDateUtilities.currentDateTime().strftime("%Y-%m-%d")
            # if start == end:
                # If we send start and end dates for intraday, it comes back with empty dataframe
            start = None
            end = None
            # if duration == "1m" and period == "1d":
            #     period = "5d" # Download 1m data for the last 5 days
        """
        data = None
        """
        with SuppressOutput(suppress_stdout=(not printCounter), suppress_stderr=(not printCounter)):
            try:
                if yfVersion == "0.2.28":
                    YfData.user_agent_headers = {
                        'User-Agent': random.choice(USER_AGENTS)}
                if "PKDevTools_Default_Log_Level" in os.environ.keys():
                    from yfinance import utils
                    yflogger = utils.get_yf_logger()
                    yflogger.setLevel(int(os.environ.get("PKDevTools_Default_Log_Level"),logging.DEBUG))
                    yf.enable_debug_mode()
                data = yf.download(
                    tickers=stockCode,
                    period=period,
                    interval=duration,
                    proxy=proxyServer,
                    progress=False,
                    rounding = True,
                    group_by='ticker',
                    timeout=self.configManager.generalTimeout/4,
                    start=start,
                    end=end,
                    auto_adjust=True,
                    threads=len(stockCode) if not isinstance(stockCode,str) else True,
                    session=yf_session
                )
                if isinstance(stockCode,str):
                    if (data is None or data.empty):
                        for ticker in shared._ERRORS:
                            err = shared._ERRORS.get(ticker)
                            # Maybe this stock is recently listed. Let's try and fetch for the last month
                            if "YFInvalidPeriodError" in err: #and "Period \'1mo\' is invalid" not in err:
                                recommendedPeriod = period
                                if isinstance(err,YFInvalidPeriodError):
                                    recommendedPeriod = err.valid_ranges[-1]
                                else:
                                    recommendedPeriod = str(err).split("[")[1].split("]")[0].split(",")[-1].strip()
                                recommendedPeriod = recommendedPeriod.replace("'","").replace("\"","")
                                # default_logger().debug(f"Sending request again for {ticker} with period:{recommendedPeriod}")
                                data = self.fetchStockData(stockCode=ticker,period=period,duration=duration,printCounter=printCounter, start=start,end=end,proxyServer=proxyServer,)
                                return data
                            elif "YFRateLimitError" in err:
                                if attempt <= 2:
                                    default_logger().debug(f"YFRateLimitError Hit! Going for attempt : {attempt+1}")
                                    # sleep(attempt*1) # Exponential backoff
                                    # return self.fetchStockData(stockCode=stockCode,period=period,duration=duration,printCounter=printCounter, start=start,end=end,screenResultsCounter=screenResultsCounter,screenCounter=screenCounter,totalSymbols=totalSymbols,exchangeSuffix=exchangeSuffix,attempt=attempt+1)
                    else:
                        multiIndex = data.keys()
                        if isinstance(multiIndex, pd.MultiIndex):
                            # If we requested for multiple stocks from yfinance
                            # we'd have received a multiindex dataframe
                            listStockCodes = multiIndex.get_level_values(0)
                            data = data.get(listStockCodes[0])
                # else:
                #     if (data is None or data.empty):
                #         if len(shared._ERRORS) > 0:
                #             default_logger().debug(shared._ERRORS)
                #         for ticker in shared._ERRORS:
                #             err = shared._ERRORS.get(ticker)
                #             if "YFRateLimitError" in err:
                #                 if attempt <= 2:
                #                     default_logger().debug(f"YFRateLimitError Hit! Going for attempt : {attempt+1}")
                                    # sleep(attempt*2) # Exponential backoff
                                    # return self.fetchStockData(stockCode=stockCode,period=period,duration=duration,printCounter=printCounter, start=start,end=end,screenResultsCounter=screenResultsCounter,screenCounter=screenCounter,totalSymbols=totalSymbols,exchangeSuffix=exchangeSuffix,attempt=attempt+1)
            except (KeyError,YFPricesMissingError) as e: # pragma: no cover
                default_logger().debug(e,exc_info=True)
                pass
            except YFRateLimitError as e:
                default_logger().debug(f"YFRateLimitError Hit! \n{e}")
                # if attempt <= 2:
                #     default_logger().debug(f"YFRateLimitError Hit! Going for attempt : {attempt+1}")
                    # sleep(attempt*2) # Exponential backoff
                    # return self.fetchStockData(stockCode=stockCode,period=period,duration=duration,printCounter=printCounter, start=start,end=end,screenResultsCounter=screenResultsCounter,screenCounter=screenCounter,totalSymbols=totalSymbols,exchangeSuffix=exchangeSuffix,attempt=attempt+1)
                pass
            except (YFInvalidPeriodError,Exception) as e: # pragma: no cover
                default_logger().debug(e,exc_info=True)                    
        if printCounter and type(screenCounter) != int:
            sys.stdout.write("\r\033[K")
            try:
                OutputControls().printOutput(
                    colorText.GREEN
                    + (
                        "[%d%%] Screened %d, Found %d. Fetching data & Analyzing %s..."
                        % (
                            int((screenCounter.value / totalSymbols) * 100),
                            screenCounter.value,
                            screenResultsCounter.value,
                            stockCode,
                        )
                    )
                    + colorText.END,
                    end="",
                )
            except ZeroDivisionError as e: # pragma: no cover
                default_logger().debug(e, exc_info=True)
                pass
            except KeyboardInterrupt: # pragma: no cover
                raise KeyboardInterrupt
            except Exception as e:  # pragma: no cover
                default_logger().debug(e, exc_info=True)
                pass
        if (data is None or len(data) == 0) and printCounter:
            OutputControls().printOutput(
                colorText.FAIL
                + "=> Failed to fetch!"
                + colorText.END,
                end="\r",
                flush=True,
            )
            raise StockDataEmptyException
        if printCounter:
            OutputControls().printOutput(
                colorText.GREEN + "=> Done!" + colorText.END,
                end="\r",
                flush=True,
            )
        """
        return data

    # Get Daily Nifty 50 Index:
    def fetchLatestNiftyDaily(self, proxyServer=None):
        return None
        data = yf.download(
            tickers="^NSEI",
            period="5d",
            interval="1d",
            proxy=proxyServer,
            progress=False,
            timeout=self.configManager.longTimeout,
        )
        return data

    # Get Data for Five EMA strategy
    def fetchFiveEmaData(self, proxyServer=None):
        return None
        nifty_sell = yf.download(
            tickers="^NSEI",
            period="5d",
            interval="5m",
            proxy=proxyServer,
            progress=False,
            timeout=self.configManager.longTimeout,
        )
        banknifty_sell = yf.download(
            tickers="^NSEBANK",
            period="5d",
            interval="5m",
            proxy=proxyServer,
            progress=False,
            timeout=self.configManager.longTimeout,
        )
        nifty_buy = yf.download(
            tickers="^NSEI",
            period="5d",
            interval="15m",
            proxy=proxyServer,
            progress=False,
            timeout=self.configManager.longTimeout,
        )
        banknifty_buy = yf.download(
            tickers="^NSEBANK",
            period="5d",
            interval="15m",
            proxy=proxyServer,
            progress=False,
            timeout=self.configManager.longTimeout,
        )
        return nifty_buy, banknifty_buy, nifty_sell, banknifty_sell

    # Load stockCodes from the watchlist.xlsx
    def fetchWatchlist(self):
        createTemplate = False
        data = pd.DataFrame()
        try:
            data = pd.read_excel("watchlist.xlsx")
        except FileNotFoundError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().printOutput(
                colorText.FAIL
                + f"  [+] watchlist.xlsx not found in {os.getcwd()}"
                + colorText.END
            )
            createTemplate = True
        try:
            if not createTemplate:
                data = data["Stock Code"].values.tolist()
        except KeyError as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().printOutput(
                colorText.FAIL
                + '  [+] Bad Watchlist Format: First Column (A1) should have Header named "Stock Code"'
                + colorText.END
            )
            createTemplate = True
        if createTemplate:
            sample = {"Stock Code": ["SBIN", "INFY", "TATAMOTORS", "ITC"]}
            sample_data = pd.DataFrame(sample, columns=["Stock Code"])
            sample_data.to_excel("watchlist_template.xlsx", index=False, header=True)
            OutputControls().printOutput(
                colorText.BLUE
                + f"  [+] watchlist_template.xlsx created in {os.getcwd()} as a referance template."
                + colorText.END
            )
            return None
        return data



# -*- coding: utf-8 -*-
# """
#     The MIT License (MIT)

#     Copyright (c) 2023 pkjmesra

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

# """
"""
Spyder Editor

This is a temporary script file.

python setup.py clean build sdist bdist_wheel

"""
import platform
import subprocess
import sys
import os
import shutil
from distutils.core import setup

import setuptools  # noqa

try:
    from pkscreener.classes import VERSION
except:
    VERSION = "0.45"
    pass

__USERNAME__ = "pkjmesra"
__PACKAGENAME__ = "pkscreener"
install_requires=[]
if os.path.exists("README.md") and os.path.isfile("README.md"):
    with open("README.md", "r") as fh:
        long_description = fh.read()
if os.path.exists("requirements.txt") and os.path.isfile("requirements.txt"):
    with open("requirements.txt", "r") as fh:
        install_requires = fh.read().splitlines()
        install_requires.append("advanced_ta")
elif os.path.exists(os.path.join(__PACKAGENAME__,"requirements.txt")) and os.path.isfile(os.path.join(__PACKAGENAME__,"requirements.txt")):
    with open(os.path.join(__PACKAGENAME__,"requirements.txt"), "r") as fh:
        install_requires = fh.read().splitlines()
        install_requires.append("advanced_ta")

talibWindowsFile = ".github/dependencies/ta_lib-0.6.0-cp312-cp312-win_amd64.whl"
talibLinuxFile = ".github/dependencies/build_tools/github/talib.sh"
if "Windows" in platform.system() and os.path.isfile(talibWindowsFile) and os.path.isfile(talibWindowsFile):
    install_requires = [talibWindowsFile].extend(install_requires)
elif "Linux" in platform.system() and os.path.isfile(talibLinuxFile) and os.path.isfile(talibLinuxFile):
    subprocess.Popen(["chmod", "+x", talibLinuxFile])
    subprocess.Popen(talibLinuxFile, shell=True)
elif "Darwin" in platform.system():
    subprocess.Popen("brew install ta-lib && brew upgrade ta-lib", shell=True)
# For Darwin, brew install ta-lib will work

SYS_MAJOR_VERSION = str(sys.version_info.major)
SYS_VERSION = SYS_MAJOR_VERSION + "." + str(sys.version_info.minor)

WHEEL_NAME = (
    __PACKAGENAME__ + "-" + VERSION + "-py" + SYS_MAJOR_VERSION + "-none-any.whl"
)
TAR_FILE = __PACKAGENAME__ + "-" + VERSION + ".tar.gz"
EGG_FILE = __PACKAGENAME__ + "-" + VERSION + "-py" + SYS_VERSION + ".egg"
DIST_FILES = [WHEEL_NAME, TAR_FILE, EGG_FILE]
DIST_DIR = "dist/"

# def _post_build():
# 	if "bdist_wheel" in sys.argv:
# 		for count, filename in enumerate(os.listdir(DIST_DIR)):
# 			if filename in DIST_FILES:
# 				os.rename(DIST_DIR + filename, DIST_DIR + filename.replace(__PACKAGENAME__+'-', __PACKAGENAME__+'_'+__USERNAME__+'-'))

# atexit.register(_post_build)
PYTHON_VERSION = (3, 12)

# def _post_build():
# 	if "bdist_wheel" in sys.argv:
# 		for count, filename in enumerate(os.listdir(DIST_DIR)):
# 			if filename in DIST_FILES:
# 				os.rename(DIST_DIR + filename, DIST_DIR + filename.replace(__PACKAGENAME__+'-', __PACKAGENAME__+'_'+__USERNAME__+'-'))

# atexit.register(_post_build)

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False
except ImportError:
    bdist_wheel = None

package_files_To_Install = ["LICENSE","README.md","requirements.txt",f"docs{os.sep}LICENSE-Others",
                            f"Disclaimer.txt",f"screenshots{os.sep}logos{os.sep}LogoWM.png"]
package_files = [__PACKAGENAME__ + ".ini","courbd.ttf"]
package_dir = os.path.join(os.getcwd(),__PACKAGENAME__)
if os.path.exists(package_dir):
    for file in package_files_To_Install:
        targetFileName = file.split(os.sep)[-1].split(".")[0] + ".txt"
        package_files.append(targetFileName)
        srcFile = os.path.join(os.getcwd(),file)
        if os.path.isfile(srcFile):
            shutil.copy(srcFile,os.path.join(package_dir,targetFileName))

setup(
    name=__PACKAGENAME__,
    packages=setuptools.find_packages(where=".", exclude=["docs", "test"]),
    cmdclass={'bdist_wheel': bdist_wheel},
    include_package_data=True,  # include everything in source control
    package_data={
        __PACKAGENAME__: package_files
    },
    # ...but exclude README.txt from all packages
    exclude_package_data={"": ["*.yml"]},
    version=VERSION,
    description="A Python-based stock screener for NSE, India with alerts to Telegram Channel (pkscreener)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=__USERNAME__,
    author_email=__USERNAME__ + "@gmail.com",
    license="OSI Approved (MIT)",
    url="https://github.com/"
    + __USERNAME__
    + "/"
    + __PACKAGENAME__,  # use the URL to the github repo
    zip_safe=False,
    entry_points="""
	[console_scripts]
	pkscreener=pkscreener.pkscreenercli:pkscreenercli
	pkbot=pkscreener.pkscreenerbot:runpkscreenerbot
	""",
    download_url="https://github.com/"
    + __USERNAME__
    + "/"
    + __PACKAGENAME__
    + "/archive/v"
    + VERSION
    + ".zip",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
    ],
    install_requires=install_requires,
    keywords=["NSE", "Technical Indicators", "Scanning", "Stock Scanners"],
),
python_requires = (">=3.12",)



#!/usr/bin/env python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import click
import platform
# # SOURCE: https://docs.python.org/2/library/curses.html
# # SOURCE: https://docs.python.org/3/howto/curses.html

# # For Windows: pip install windows-curses
# import curses
# window = curses.initscr() # Initialize the library. Returns a WindowObject which represents the whole screen.
# window.keypad(True) # Escape sequences generated by some keys (keypad, function keys) will be interpreted by curses.
# curses.cbreak() # Keys are read one by one. Also safer than curses.raw() because you can interrupt a running script with SIGINT (Ctrl + C).
# curses.noecho() # Prevent getch() keys from being visible when pressed. Echoing of input characters is turned off.

# # Initialize colors.
# curses.start_color() # Must be called if you want to use colors.
# curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_WHITE)
# curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_BLACK)
# curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)
# curses.init_pair(4, curses.COLOR_GREEN, curses.COLOR_BLACK)
# curses.init_pair(5, curses.COLOR_YELLOW, curses.COLOR_BLACK)
# curses.init_pair(6, curses.COLOR_BLUE, curses.COLOR_BLACK)
# curses.init_pair(7, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
# curses.init_pair(8, curses.COLOR_CYAN, curses.COLOR_BLACK)
# black = curses.color_pair(1)
# white = curses.color_pair(2)
# red = curses.color_pair(3)
# green = curses.color_pair(4)
# yellow = curses.color_pair(5)
# blue = curses.color_pair(6)
# magenta = curses.color_pair(7)
# cyan = curses.color_pair(8)

# # -----

# def draw_menu(menuItems, selectedIndex):
#     #  window.erase()
#     window.clear()

#     # Print a vertical menu.
#     line = 1
#     for i in range(len(menuItems)):
#         window.addstr(' ')
#         newLines = ""
#         menu = menuItems[i]
#         menuText = menu.keyTextLabel()
#         if line != menu.line:
#           window.addstr('\n')
#           line += 1
#         window.addstr(f"{menuText}\n", black if i == selectedIndex else white)
#         #   line += 1
#         line += 1

#     # # Print a dividing line.
#     # window.addstr(('-' * 80) + '\n')

#     # # Print a horizontal menu.
#     # for i in range(len(menuItems)):
#     #     window.addstr(' ')
#     #     window.addstr(menuItems[i], black if i == selectedIndex else white)

#     # window.addstr('\n')

# # -----

# def process_input(menuItems, selectedIndex):
#     userInput = window.getch()

#     if userInput == curses.KEY_LEFT or userInput == curses.KEY_UP:
#         # Loop around backwards.
#         selectedIndex = (selectedIndex - 1 + len(menuItems)) % len(menuItems)

#     elif userInput == curses.KEY_RIGHT or userInput == curses.KEY_DOWN:
#         # Loop around forwards.
#         selectedIndex = (selectedIndex + 1) % len(menuItems)

#     # If curses.nonl() is called, Enter key = \r else \n.
#     elif userInput == curses.KEY_ENTER or chr(userInput) in '\r\n':
#         # If the last option, exit, is selected.
#         if selectedIndex == len(menuItems) - 1:
#             wait_for_any_keypress()
#             curses.endwin() # De-initialize the library, and return terminal to normal status.    <-- Works without this on Windows, however in Linux you can't type in the terminal after exiting without this :P
#             exit(0)

#         window.addstr('\n Selected index: {}\n'.format(selectedIndex))
#         # wait_for_any_keypress()

#     else:
#         window.addstr("\n The pressed key '{}' {} is not associated with a menu function.\n".format(chr(userInput), userInput))
#         # wait_for_any_keypress()

#     return selectedIndex

# # -----

# def wait_for_any_keypress():
#     window.addstr('\n Press any key to continue . . . ')
#     window.getch()

# # -----

# def main():
#     selectedIndex = 0
#     while True:
#         draw_menu(MENU_ITEMS, selectedIndex)
#         selectedIndex = process_input(MENU_ITEMS, selectedIndex)

# from pkscreener.classes.MenuOptions import menus
# m = menus()

# MENU_ITEMS = m.renderForMenu(asList=True)
# #[
# #     ' Option 1 ',
# #     ' Option 2 ',
# #     ' Option 3 ',
# #     ' Exit ',
# # ]

# if __name__ == '__main__':
#     main()

def getKeyBoardArrowInput(message="Use Left / Right arrow keys to slide (going back / forward) the time-window!"):
    printable = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
    direction = None
    if message is not None and len(message) > 0:
        click.echo(message, nl=False)
    c = click.getchar()
    # click.echo()
    supportedDirections = {'\x1b[A':'UP','\x1b[B':'DOWN','\x1b[C':'RIGHT','\x1b[D':'LEFT','àK':'LEFT','àH':'UP','àP':'DOWN','àM':'RIGHT','0x0d':'RETURN','\r':'RETURN','\n':'RETURN','c':'CANCEL','C':'CANCEL'}
    if c in supportedDirections.keys():
        # click.echo('Invalid input :(')
        # click.echo('You pressed: "' + ''.join([ '\\'+hex(ord(i))[1:] if i not in printable else i for i in c ]) +'"' )
        direction = supportedDirections.get(c)
    else:
        if len(c) >= 2: # Windows/Mac may return 2 character key for arrow
            if "Windows" in platform.system():
                supportedWinDirections = {"K":"LEFT", "H": "UP", "P" : "DOWN", "M": "RIGHT"}
                if c[-1] in supportedWinDirections.keys():
                    direction = supportedWinDirections.get(c[-1])
            else:
                supportedOthDirections = {"D":"LEFT", "A": "UP", "B" : "DOWN", "C": "RIGHT"}
                if c[-1] in supportedOthDirections.keys():
                    direction = supportedOthDirections.get(c[-1])

    return direction



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
from PKDevTools.classes.Environment import PKEnvironment
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.PKDateUtilities import PKDateUtilities

import pkscreener.classes.ConfigManager as ConfigManager
from pkscreener.classes.Fetcher import screenerStockDataFetcher

configManager = ConfigManager.tools()

def run_workflow(command=None, user=None, options=None, workflowType="B",repo=None,owner=None,branch=None,ghp_token=None,workflow_name=None,workflow_postData=None):
    if owner is None:
        owner = os.popen('git ls-remote --get-url origin | cut -d/ -f4').read().replace("\n","")
    if repo is None:
        repo = os.popen('git ls-remote --get-url origin | cut -d/ -f5').read().replace(".git","").replace("\n","")
    if branch is None:
        branch = "main"
    timestamp = int(PKDateUtilities.currentDateTimestamp())
    if workflowType == "B":
        if workflow_name is None:
            workflow_name = "w13-workflow-backtest_generic.yml"
        options = f'{options.replace("B:","")}:D:D:D:D:D'.replace("::",":")
        data = (
            '{"ref":"'
            + branch
            + '","inputs":{"user":"'
            + f"{user}"
            + '","params":"'
            + f"{options}"
            + '","name":"'
            + f"{command}"
            + '"}}'
        )
    elif workflowType == "X" or workflowType == "G" or workflowType == "P":
        if workflow_name is None:
            workflow_name = "w8-workflow-alert-scan_generic.yml"
        if user is None or len(user) == 0:
            user = ""
            data = (
                '{"ref":"'
                + branch
                + '","inputs":{"user":"'
                + f"{user}"
                + '","params":"'
                + f'-a Y -e --triggertimestamp {timestamp} -p -o {options.replace("_",":")}:D:D:D:D:D'.replace("::",":")
                + '","ref":"main"}}'
            )
        else:
            data = (
                '{"ref":"'
                + branch
                + '","inputs":{"user":"'
                + f"{user}"
                + '","params":"'
                + f'-a Y -e --triggertimestamp {timestamp} -p -u {user} -o {options.replace("_",":")}:D:D:D:D:D'.replace("::",":")
                + '","ref":"main"}}'
            )
    elif workflowType == "R": #Restart bot
        if workflow_name is None:
            workflow_name = "w3-workflow-bot.yml"
        data = (
                '{"ref":"'
                + branch
                + '","inputs":{"branch-name":"main","cliOptions":""}}'
            )
    elif workflowType == "O": #Others
        if workflow_name is None or workflow_postData is None or ghp_token is None:
            raise Exception("workflow_name, workflow_postData, and ghp_token must NOT be blank!")
        data = workflow_postData
    elif workflowType == "S": # Scanner job kick off for 1-on-1 alerts
        cmd_options = options.replace("_",":")
        if workflow_name is None:
            workflow_name = "w8-workflow-alert-scan_generic.yml"
        if 'ALERT_TRIGGER' in os.environ.keys() and os.environ["ALERT_TRIGGER"] == 'Y':
            alertTrigger = 'Y'
        else:
            alertTrigger = 'N'
        if user is None or len(user) == 0:
            user = ""
        data = (
                    '{"ref":"'
                    + branch
                    + '","inputs":{"user":"'
                    + f"{user}"
                    + '","params":"'
                    + f'{cmd_options} --triggertimestamp {timestamp}'
                    + f'","ref":"{branch}","alertTrigger":"'
                    + f"{alertTrigger}"
                    + '","name":"'
                    + f"{command}"
                    + '"}}'
                )

    if ghp_token is None:
        _, _, _, ghp_token = PKEnvironment().secrets
    url = f"https://api.github.com/repos/{owner}/{repo}/actions/workflows/{workflow_name}/dispatches"

    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {ghp_token}",
        "Content-Type": "application/json",
    }

    fetcher = screenerStockDataFetcher(configManager)
    resp = fetcher.postURL(url, data=data, headers=headers)
    if resp.status_code == 204:
        OutputControls().printOutput(f"Workflow {workflow_name} Triggered!")
    else:
        OutputControls().printOutput(f"Something went wrong while triggering {workflow_name}")
    return resp


# resp = run_workflow("B_12_1","-1001785195297","B:12:1")
